"""
Reports are configurable summaries after Robocop scan. For example, it can be a total number of issues discovered.
They are dynamically loaded during setup according to a command line configuration.

Each report class collects rules messages from linter and parses it. At the end of the scan it will print the report.

To enable report use ``-r`` / ``--report`` argument and the name of the report.
You can use separate arguments (``-r report1 -r report2``) or comma-separated list (``-r report1,report2``). Example::

    robocop --report rules_by_id,some_other_report path/to/file.robot

To enable all reports use ``--report all``.
"""
from collections import defaultdict
from operator import itemgetter
from timeit import default_timer as timer

import robocop.exceptions
from robocop.rules import Message


class Report:
    def configure(self, name, value):
        raise robocop.exceptions.ConfigGeneralError(
            f"Provided param '{name}' for report '{getattr(self, 'name')}' does not exist"
        )  # noqa


class RulesByIdReport(Report):
    """
    Report name: ``rules_by_id``

    Report that groups linter rules messages by rule id and prints it ordered by most common message.
    Example::

        Issues by ids:
        W0502 (too-little-calls-in-keyword) : 5
        W0201 (missing-doc-keyword)         : 4
        E0401 (parsing-error)               : 3
        W0301 (not-allowed-char-in-name)    : 2
        E0901 (keyword-after-return)        : 1
    """

    def __init__(self):
        self.name = "rules_by_id"
        self.description = "Groups detected issues by rule id and prints it ordered by most common"
        self.message_counter = defaultdict(int)

    def add_message(self, message: Message):  # noqa
        self.message_counter[message.get_fullname()] += 1

    def get_report(self) -> str:
        message_counter_ordered = sorted(self.message_counter.items(), key=itemgetter(1), reverse=True)
        report = "\nIssues by IDs:\n"
        if not message_counter_ordered:
            report += "No issues found"
            return report
        longest_name = max(len(msg[0]) for msg in message_counter_ordered)
        report += "\n".join(f"{message:{longest_name}} : {count}" for message, count in message_counter_ordered)
        return report


class RulesBySeverityReport(Report):
    """
    Report name: ``rules_by_error_type``

    Report that groups linter rules messages by severity and prints total of issues per every severity level.

    Example::

        Found 15 issues: 11 WARNING(s), 4 ERROR(s).
    """

    def __init__(self):
        self.name = "rules_by_error_type"
        self.description = "Prints total number of issues grouped by severity"
        self.severity_counter = defaultdict(int)

    def add_message(self, message: Message):
        self.severity_counter[message.severity] += 1

    def get_report(self) -> str:
        issues_count = sum(self.severity_counter.values())
        if not issues_count:
            return "\nFound 0 issues"
        report = f"\nFound {issues_count} issue(s): "
        report += ", ".join(f"{count} {severity.name}(s)" for severity, count in self.severity_counter.items())
        report += "."
        return report


class ReturnStatusReport(Report):
    """
    Report name: ``return_status``

    Report that checks if number of returned rules messages for given severity value does not exceed preset threshold.
    That information is later used as a return status from Robocop.
    """

    def __init__(self):
        self.name = "return_status"
        self.description = "Checks if number of specific issues exceed quality gate limits"
        self.return_status = 0
        self.counter = RulesBySeverityReport()
        self.quality_gate = {"E": 0, "W": 0, "I": -1}

    def configure(self, name, value):
        if name not in ["quality_gate", "quality_gates"]:
            super().configure(name, value)
        for val in value.split(":"):
            try:
                name, count = val.split("=", maxsplit=1)
                if name.upper() in self.quality_gate:
                    self.quality_gate[name.upper()] = int(count)
            except ValueError:
                continue

    def add_message(self, message: Message):
        self.counter.add_message(message)

    def get_report(self) -> str:
        for severity, count in self.counter.severity_counter.items():
            threshold = self.quality_gate.get(severity.value, 0)
            if -1 < threshold < count:
                self.return_status += count - threshold
        self.return_status = min(self.return_status, 255)


class TimeTakenReport(Report):
    """
    Report name: ``scan_timer``

    Report that returns Robocop execution time
    """

    def __init__(self):
        self.name = "scan_timer"
        self.description = "Returns Robocop execution time"
        self.start_time = timer()

    def add_message(self, *args):
        pass

    def get_report(self) -> str:
        return f"\nScan took {timer() - self.start_time:.3f}s"


class JsonReport(Report):
    """
    Report name: ``json_report``

    Report that returns list of found issues in JSON format.
    """

    def __init__(self):
        self.name = "json_report"
        self.description = "Accumulates found issues in JSON format"
        self.issues = []

    def add_message(self, message: Message):
        self.issues.append(message.to_json())

    def get_report(self):
        return None


class FileStatsReport(Report):
    """
    Report name: ``file_stats``

    Report that displays overall statistics about number of processed files.
    """

    def __init__(self):
        self.name = "file_stats"
        self.description = "Prints overall statistics about number of processed files"
        self.files_count = 0
        self.files_with_issues = set()

    def add_message(self, message: Message):
        self.files_with_issues.add(message.source)

    def get_report(self) -> str:
        if not self.files_count:
            return "\nNo files were processed"
        if not self.files_with_issues:
            return f"\nProcessed {self.files_count} file(s) but no issues were found"

        return (
            f"\nProcessed {self.files_count} file(s) from which {len(self.files_with_issues)} "
            f"file(s) contained issues"
        )



#Note: code gotten from _pydev_imports_tipper.

import sys

def _imp(name, log=None):
    try:
        return __import__(name)
    except:
        if '.' in name:
            sub = name[0:name.rfind('.')]
            
            if log is not None:
                log.add_content('Unable to import', name, 'trying with', sub)
                log.add_exception()
            
            return _imp(sub, log)
        else:
            s = 'Unable to import module: %s - sys.path: %s' % (str(name), sys.path)
            if log is not None:
                log.add_content(s)
                log.add_exception()
            
            raise ImportError(s)
        

IS_IPY = False
if sys.platform == 'cli':
    IS_IPY = True
    _old_imp = _imp
    def _imp(name, log=None):
        #We must add a reference in clr for .Net
        import clr #@UnresolvedImport
        initial_name = name
        while '.' in name:
            try:
                clr.AddReference(name)
                break #If it worked, that's OK.
            except:
                name = name[0:name.rfind('.')]
        else:
            try:
                clr.AddReference(name)
            except:
                pass #That's OK (not dot net module).
        
        return _old_imp(initial_name, log)
    

def import_name(name, log=None):
    mod = _imp(name, log)

    components = name.split('.')

    old_comp = None
    for comp in components[1:]:
        try:
            #this happens in the following case:
            #we have mx.DateTime.mxDateTime.mxDateTime.pyd
            #but after importing it, mx.DateTime.mxDateTime shadows access to mxDateTime.pyd
            mod = getattr(mod, comp)
        except AttributeError:
            if old_comp != comp:
                raise
        
        old_comp = comp
        
    return mod
    



from robot.api.parsing import (
    ModelTransformer,
    Token,
    EmptyLine,
    Comment,
    ModelVisitor,
    ForHeader,
    End,
    IfHeader,
    ElseHeader,
    ElseIfHeader,
)

from robotidy.decorators import check_start_end_line
from robotidy.utils import round_to_four, is_suite_templated


class AlignTestCases(ModelTransformer):
    """
    Align Test Cases to columns.

    Currently only templated tests are supported. Following code:

        *** Test Cases ***    baz    qux
        # some comment
        test1    hi    hello
        test2 long test name    asdfasdf    asdsdfgsdfg

    will be transformed to:

        *** Test Cases ***      baz         qux
        # some comment
        test1                   hi          hello
        test2 long test name    asdfasdf    asdsdfgsdfg
                                bar1        bar2

    If you don't want to align test case section that does not contain header names (in above example baz and quz are
    header names) then configure `only_with_headers` parameter:

        robotidy -c AlignSettingsSection:only_with_hedaers:True <src>

    Supports global formatting params: ``--startline``, ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/AlignTestCases.html for more examples.
    """

    ENABLED = False

    def __init__(self, only_with_headers: bool = False, min_width: int = None):
        self.only_with_headers = only_with_headers
        self.min_width = min_width
        self.widths = None
        self.test_name_len = 0
        self.name_line = 0
        self.indent = 0

    def visit_File(self, node):  # noqa
        if not is_suite_templated(node):
            return node
        return self.generic_visit(node)

    def visit_If(self, node):  # noqa
        self.indent += 1
        self.generic_visit(node)
        self.indent -= 1
        return node

    visit_Else = visit_ElseIf = visit_For = visit_If

    @check_start_end_line
    def visit_TestCaseSection(self, node):  # noqa
        if len(node.header.data_tokens) == 1 and self.only_with_headers:
            return node
        counter = ColumnWidthCounter(self.formatting_config)
        counter.visit(node)
        self.widths = counter.widths
        return self.generic_visit(node)

    @check_start_end_line
    def visit_Statement(self, statement):  # noqa
        if statement.type == Token.TESTCASE_NAME:
            self.test_name_len = len(statement.tokens[0].value)
            self.name_line = statement.lineno
        elif statement.type == Token.TESTCASE_HEADER:
            self.align_header(statement)
        elif not isinstance(
            statement,
            (Comment, EmptyLine, ForHeader, IfHeader, ElseHeader, ElseIfHeader, End),
        ):
            self.align_statement(statement)
        return statement

    def align_header(self, statement):
        tokens = []
        # *** Test Cases ***            baz                            qux
        # *** Test Cases ***      baz         qux
        for index, token in enumerate(statement.data_tokens[:-1]):
            tokens.append(token)
            if self.min_width:
                separator = max(self.formatting_config.space_count, self.min_width - len(token.value)) * " "
            else:
                separator = (self.widths[index] - len(token.value) + self.formatting_config.space_count) * " "
            tokens.append(Token(Token.SEPARATOR, separator))
        tokens.append(statement.data_tokens[-1])
        tokens.append(statement.tokens[-1])  # eol
        statement.tokens = tokens
        return statement

    def align_statement(self, statement):
        tokens = []
        for line in statement.lines:
            strip_line = [t for t in line if t.type not in (Token.SEPARATOR, Token.EOL)]
            line_pos = 0
            exp_pos = 0
            widths = self.get_widths(statement)
            for token, width in zip(strip_line, widths):
                if self.min_width:
                    exp_pos += max(width + self.formatting_config.space_count, self.min_width)
                else:
                    exp_pos += width + self.formatting_config.space_count
                if self.test_name_len:
                    if self.name_line == statement.lineno:
                        exp_pos -= self.test_name_len
                    self.test_name_len = 0
                tokens.append(Token(Token.SEPARATOR, (exp_pos - line_pos) * " "))
                tokens.append(token)
                line_pos += len(token.value) + exp_pos - line_pos
            tokens.append(line[-1])
        statement.tokens = tokens

    def get_widths(self, statement):
        indent = self.indent
        if isinstance(statement, (ForHeader, End, IfHeader, ElseHeader, ElseIfHeader)):
            indent -= 1
        if not indent:
            return self.widths
        return [max(width, indent * self.formatting_config.space_count) for width in self.widths]

    def visit_SettingSection(self, node):  # noqa
        return node

    visit_VariableSection = visit_KeywordSection = visit_CommentSection = visit_SettingSection


class ColumnWidthCounter(ModelVisitor):
    def __init__(self, formatting_config):
        self.widths = []
        self.formatting_config = formatting_config
        self.test_name_lineno = -1
        self.any_one_line_test = False
        self.header_with_cols = False

    def visit_TestCaseSection(self, node):  # noqa
        self.generic_visit(node)
        if not self.header_with_cols and not self.any_one_line_test and self.widths:
            self.widths[0] = 0
        self.widths = [round_to_four(length) for length in self.widths]

    @check_start_end_line
    def visit_Statement(self, statement):  # noqa
        if statement.type == Token.COMMENT:
            return
        if statement.type == Token.TESTCASE_HEADER:
            if len(statement.data_tokens) > 1:
                self.header_with_cols = True
                self._count_widths_from_statement(statement)
        elif statement.type == Token.TESTCASE_NAME:
            if self.widths:
                self.widths[0] = max(self.widths[0], len(statement.name))
            else:
                self.widths.append(len(statement.name))
            self.test_name_lineno = statement.lineno
        else:
            if self.test_name_lineno == statement.lineno:
                self.any_one_line_test = True
            if not isinstance(statement, (ForHeader, IfHeader, ElseHeader, ElseIfHeader, End)):
                self._count_widths_from_statement(statement, indent=1)

    def _count_widths_from_statement(self, statement, indent=0):
        for line in statement.lines:
            line = [t for t in line if t.type not in (Token.SEPARATOR, Token.EOL)]
            for index, token in enumerate(line, start=indent):
                if index < len(self.widths):
                    self.widths[index] = max(self.widths[index], len(token.value))
                else:
                    self.widths.append(len(token.value))



'''
Entry point module (keep at root):

Used to run with tests with unittest/pytest/nose.
'''

import os


def main():
    import sys

    # Separate the nose params and the pydev params.
    pydev_params = []
    other_test_framework_params = []
    found_other_test_framework_param = None

    NOSE_PARAMS = '--nose-params'
    PY_TEST_PARAMS = '--py-test-params'

    for arg in sys.argv[1:]:
        if not found_other_test_framework_param and arg != NOSE_PARAMS and arg != PY_TEST_PARAMS:
            pydev_params.append(arg)

        else:
            if not found_other_test_framework_param:
                found_other_test_framework_param = arg
            else:
                other_test_framework_params.append(arg)

    try:
        # Convert to the case stored in the filesystem
        import win32api

        def get_with_filesystem_case(f):
            return win32api.GetLongPathName(win32api.GetShortPathName(f))

    except:

        def get_with_filesystem_case(f):
            return f

    # Here we'll run either with nose or with the pydev_runfiles.
    from _pydev_runfiles import pydev_runfiles
    from _pydev_runfiles import pydev_runfiles_xml_rpc
    from _pydevd_bundle import pydevd_constants
    from pydevd_file_utils import canonical_normalized_path

    DEBUG = 0
    if DEBUG:
        sys.stdout.write('Received parameters: %s\n' % (sys.argv,))
        sys.stdout.write('Params for pydev: %s\n' % (pydev_params,))
        if found_other_test_framework_param:
            sys.stdout.write('Params for test framework: %s, %s\n' % (found_other_test_framework_param, other_test_framework_params))

    try:
        configuration = pydev_runfiles.parse_cmdline([sys.argv[0]] + pydev_params)
    except:
        sys.stderr.write('Command line received: %s\n' % (sys.argv,))
        raise
    pydev_runfiles_xml_rpc.initialize_server(configuration.port)  # Note that if the port is None, a Null server will be initialized.

    NOSE_FRAMEWORK = "nose"
    PY_TEST_FRAMEWORK = "py.test"
    test_framework = None  # Default (pydev)
    try:
        if found_other_test_framework_param:
            if found_other_test_framework_param == NOSE_PARAMS:
                test_framework = NOSE_FRAMEWORK
                import nose

            elif found_other_test_framework_param == PY_TEST_PARAMS:
                test_framework = PY_TEST_FRAMEWORK
                import pytest

            else:
                raise ImportError('Test framework: %s not supported.' % (found_other_test_framework_param,))

        else:
            raise ImportError()

    except ImportError:
        if found_other_test_framework_param:
            raise

        test_framework = None

    # Clear any exception that may be there so that clients don't see it.
    # See: https://sourceforge.net/tracker/?func=detail&aid=3408057&group_id=85796&atid=577329
    if hasattr(sys, 'exc_clear'):
        sys.exc_clear()

    if not test_framework:

        return pydev_runfiles.main(configuration)  # Note: still doesn't return a proper value.

    else:
        # We'll convert the parameters to what nose or py.test expects.
        # The supported parameters are:
        # runfiles.py  --config-file|-t|--tests <Test.test1,Test2>  dirs|files --nose-params xxx yyy zzz
        # (all after --nose-params should be passed directly to nose)

        # In java:
        # --tests = Constants.ATTR_UNITTEST_TESTS
        # --config-file = Constants.ATTR_UNITTEST_CONFIGURATION_FILE

        # The only thing actually handled here are the tests that we want to run, which we'll
        # handle and pass as what the test framework expects.

        py_test_accept_filter = {}
        files_to_tests = configuration.files_to_tests

        if files_to_tests:
            # Handling through the file contents (file where each line is a test)
            files_or_dirs = []
            for file, tests in files_to_tests.items():
                if test_framework == NOSE_FRAMEWORK:
                    for test in tests:
                        files_or_dirs.append(file + ':' + test)

                elif test_framework == PY_TEST_FRAMEWORK:
                    py_test_accept_filter[file] = tests
                    py_test_accept_filter[canonical_normalized_path(file)] = tests
                    files_or_dirs.append(file)

                else:
                    raise AssertionError('Cannot handle test framework: %s at this point.' % (test_framework,))

        else:
            if configuration.tests:
                # Tests passed (works together with the files_or_dirs)
                files_or_dirs = []
                for file in configuration.files_or_dirs:
                    if test_framework == NOSE_FRAMEWORK:
                        for t in configuration.tests:
                            files_or_dirs.append(file + ':' + t)

                    elif test_framework == PY_TEST_FRAMEWORK:
                        py_test_accept_filter[file] = configuration.tests
                        py_test_accept_filter[canonical_normalized_path(file)] = configuration.tests
                        files_or_dirs.append(file)

                    else:
                        raise AssertionError('Cannot handle test framework: %s at this point.' % (test_framework,))
            else:
                # Only files or dirs passed (let it do the test-loading based on those paths)
                files_or_dirs = configuration.files_or_dirs

        argv = other_test_framework_params + files_or_dirs

        if test_framework == NOSE_FRAMEWORK:
            # Nose usage: http://somethingaboutorange.com/mrl/projects/nose/0.11.2/usage.html
            # show_stdout_option = ['-s']
            # processes_option = ['--processes=2']
            argv.insert(0, sys.argv[0])
            if DEBUG:
                sys.stdout.write('Final test framework args: %s\n' % (argv[1:],))

            from _pydev_runfiles import pydev_runfiles_nose
            PYDEV_NOSE_PLUGIN_SINGLETON = pydev_runfiles_nose.start_pydev_nose_plugin_singleton(configuration)
            argv.append('--with-pydevplugin')
            # Return 'not' because it will return 'success' (so, exit == 0 if success)
            return not nose.run(argv=argv, addplugins=[PYDEV_NOSE_PLUGIN_SINGLETON])

        elif test_framework == PY_TEST_FRAMEWORK:

            if '--coverage_output_dir' in pydev_params and '--coverage_include' in pydev_params:
                coverage_output_dir = pydev_params[pydev_params.index('--coverage_output_dir') + 1]
                coverage_include = pydev_params[pydev_params.index('--coverage_include') + 1]
                try:
                    import pytest_cov
                except ImportError:
                    sys.stderr.write('To do a coverage run with pytest the pytest-cov library is needed (i.e.: pip install pytest-cov).\n\n')
                    raise

                argv.insert(0, '--cov-append')
                argv.insert(1, '--cov-report=')
                argv.insert(2, '--cov=%s' % (coverage_include,))

                import time
                os.environ['COVERAGE_FILE'] = os.path.join(coverage_output_dir, '.coverage.%s' % (time.time(),))

            if DEBUG:
                sys.stdout.write('Final test framework args: %s\n' % (argv,))
                sys.stdout.write('py_test_accept_filter: %s\n' % (py_test_accept_filter,))

            def dotted(p):
                # Helper to convert path to have dots instead of slashes
                return os.path.normpath(p).replace(os.sep, "/").replace('/', '.')

            curr_dir = os.path.realpath('.')
            curr_dotted = dotted(curr_dir) + '.'

            # Overcome limitation on py.test:
            # When searching conftest if we have a structure as:
            # /my_package
            # /my_package/conftest.py
            # /my_package/tests
            # /my_package/tests/test_my_package.py
            # The test_my_package won't have access to the conftest contents from the
            # test_my_package.py file unless the working dir is set to /my_package.
            #
            # See related issue (for which we work-around below):
            # https://bitbucket.org/hpk42/pytest/issue/639/conftest-being-loaded-twice-giving

            for path in sys.path:
                path_dotted = dotted(path)
                if curr_dotted.startswith(path_dotted):
                    os.chdir(path)
                    break

            remove = []
            for i in range(len(argv)):
                arg = argv[i]
                # Workaround bug in py.test: if we pass the full path it ends up importing conftest
                # more than once (so, always work with relative paths).
                if os.path.isfile(arg) or os.path.isdir(arg):

                    # Args must be passed with the proper case in the filesystem (otherwise
                    # python itself may not recognize it).
                    arg = get_with_filesystem_case(arg)
                    argv[i] = arg

                    from os.path import relpath
                    try:
                        # May fail if on different drives
                        arg = relpath(arg)
                    except ValueError:
                        pass
                    else:
                        argv[i] = arg
                elif '<unable to get>' in arg:
                    remove.append(i)

            for i in reversed(remove):
                del argv[i]

            # To find our runfile helpers (i.e.: plugin)...
            d = os.path.dirname(__file__)
            if d not in sys.path:
                sys.path.insert(0, d)

            import pickle, zlib, base64

            # Update environment PYTHONPATH so that it finds our plugin if using xdist.
            os.environ['PYTHONPATH'] = os.pathsep.join(sys.path)

            # Set what should be skipped in the plugin through an environment variable
            s = base64.b64encode(zlib.compress(pickle.dumps(py_test_accept_filter)))
            s = s.decode('ascii')  # Must be str in py3.
            os.environ['PYDEV_PYTEST_SKIP'] = s

            # Identifies the main pid (i.e.: if it's not the main pid it has to connect back to the
            # main pid to give xml-rpc notifications).
            os.environ['PYDEV_MAIN_PID'] = str(os.getpid())
            os.environ['PYDEV_PYTEST_SERVER'] = str(configuration.port)

            argv.append('-p')
            argv.append('_pydev_runfiles.pydev_runfiles_pytest2')
            return pytest.main(argv)

        else:
            raise AssertionError('Cannot handle test framework: %s at this point.' % (test_framework,))


if __name__ == '__main__':
    try:
        main()
    finally:
        try:
            # The server is not a daemon thread, so, we have to ask for it to be killed!
            from _pydev_runfiles import pydev_runfiles_xml_rpc
            pydev_runfiles_xml_rpc.force_server_kill()
        except:
            pass  # Ignore any errors here

    import sys
    import threading
    if hasattr(sys, '_current_frames') and hasattr(threading, 'enumerate'):
        import time
        import traceback

        class DumpThreads(threading.Thread):

            def run(self):
                time.sleep(10)

                thread_id_to_name = {}
                try:
                    for t in threading.enumerate():
                        thread_id_to_name[t.ident] = '%s  (daemon: %s)' % (t.name, t.daemon)
                except:
                    pass

                stack_trace = [
                    '===============================================================================',
                    'pydev pyunit runner: Threads still found running after tests finished',
                    '================================= Thread Dump =================================']

                for thread_id, stack in sys._current_frames().items():
                    stack_trace.append('\n-------------------------------------------------------------------------------')
                    stack_trace.append(" Thread %s" % thread_id_to_name.get(thread_id, thread_id))
                    stack_trace.append('')

                    if 'self' in stack.f_locals:
                        sys.stderr.write(str(stack.f_locals['self']) + '\n')

                    for filename, lineno, name, line in traceback.extract_stack(stack):
                        stack_trace.append(' File "%s", line %d, in %s' % (filename, lineno, name))
                        if line:
                            stack_trace.append("   %s" % (line.strip()))
                stack_trace.append('\n=============================== END Thread Dump ===============================')
                sys.stderr.write('\n'.join(stack_trace))

        dump_current_frames_thread = DumpThreads()
        dump_current_frames_thread.daemon = True  # Daemon so that this thread doesn't halt it!
        dump_current_frames_thread.start()



#!/usr/bin/env python
"""Simple Gtk example to manually test event loop integration.

To run this:
1) Enable the PyDev GUI event loop integration for gtk3
2) do an execfile on this script
3) ensure you have a working GUI simultaneously with an
   interactive console
"""

if __name__ == '__main__':
    from gi.repository import Gtk
    
    
    def hello_world(wigdet, data=None):
        print("Hello World")
    
    def delete_event(widget, event, data=None):
        return False
    
    def destroy(widget, data=None):
        Gtk.main_quit()
    
    window = Gtk.Window(Gtk.WindowType.TOPLEVEL)
    window.connect("delete_event", delete_event)
    window.connect("destroy", destroy)
    button = Gtk.Button("Hello World")
    button.connect("clicked", hello_world, None)
    
    window.add(button)
    button.show()
    window.show()
    





def foo(arg):
    print('on foo mark', arg)
    return arg + 1


def main():
    # Note that we have multiple foo calls and we have to differentiate and stop at the
    # proper one.
    foo(foo(foo(foo(1))))  # break here


if __name__ == '__main__':
    main()

    print('TEST SUCEEDED')



if __name__ == '__main__':
    import sys
    import os
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))

    from not_my_code import other

    def callback2():
        other.raise_exception()

    def callback1():
        other.call_me_back2(callback2)

    other.call_me_back1(callback1)  # break here



# Original work Copyright 2018 Palantir Technologies, Inc. (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import re
from robocorp_ls_core.protocols import Sentinel

_RE_FIRST_CAP = re.compile("(.)([A-Z][a-z]+)")
_RE_ALL_CAP = re.compile("([a-z0-9])([A-Z])")


class MethodDispatcher(object):
    """JSON RPC dispatcher that calls methods on itself.

    Method names are computed by converting camel case to snake case, slashes with double underscores, and removing
    dollar signs.
    """

    def __getitem__(self, item):
        try:
            cache = self.__method_dispatcher_cache__
        except AttributeError:
            cache = self.__method_dispatcher_cache__ = {}

        handler = cache.get(item)
        if handler is None:

            method_name = f"m_{_method_to_string(item)}"
            if not hasattr(self, method_name):
                handler = Sentinel
            else:
                method = getattr(self, method_name)

                @functools.wraps(method)
                def _handler(params):
                    return method(**(params or {}))

                handler = _handler

            cache[item] = handler

        if handler is Sentinel:
            raise KeyError()
        return handler


def _method_to_string(method):
    return _camel_to_underscore(method.replace("/", "__").replace("$", ""))


def _camel_to_underscore(string):
    s1 = _RE_FIRST_CAP.sub(r"\1_\2", string)
    return _RE_ALL_CAP.sub(r"\1_\2", s1).lower()



import os
import re
import sys
import subprocess
import tempfile
import unittest


#=======================================================================================================================
# Test
#=======================================================================================================================
class Test(unittest.TestCase):
    """
    Unittest for pydev_coverage.py.
    TODO:
      - 'combine' in arguments
      - no 'combine' and no 'pydev-analyze' in arguments
    """

    def setUp(self):
        unittest.TestCase.setUp(self)
        project_path = os.path.dirname(os.path.dirname(__file__))
        self._resources_path = os.path.join(project_path, "tests_python", "resources")
        self._coverage_file = os.path.join(project_path, "pydev_coverage.py")

    def _do_analyze(self, files):
        invalid_files = []

        p = subprocess.Popen([sys.executable, self._coverage_file, "--pydev-analyze"],
                             stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
        __, stderrdata = p.communicate("|".join(files).encode())

        if stderrdata:
            match = re.search("Invalid files not passed to coverage: (.*?)$",
                              stderrdata.decode(), re.M)  # @UndefinedVariable
            if match:
                invalid_files = [f.strip() for f in match.group(1).split(",")]
        return invalid_files

    def test_pydev_analyze_ok(self):
        ref_valid_files = [__file__,
            os.path.join(self._resources_path, "_debugger_case18.py")]
        ref_invalid_files = []

        invalid_files = self._do_analyze(ref_valid_files)

        self.assertEqual(ref_invalid_files, invalid_files)

    def test_pydev_analyse_non_standard_encoding(self):
        ref_valid_files = [os.path.join(self._resources_path,
                                        "_pydev_coverage_cyrillic_encoding_py%i.py"
                                        % sys.version_info[0])]
        ref_invalid_files = []

        invalid_files = self._do_analyze(ref_valid_files + ref_invalid_files)

        self.assertEqual(ref_invalid_files, invalid_files)

    def test_pydev_analyse_invalid_files(self):
        with tempfile.NamedTemporaryFile(suffix=".pyx") as pyx_file:
            ref_valid_files = []
            ref_invalid_files = [os.path.join(self._resources_path,
                                              "_pydev_coverage_syntax_error.py"),
                                 pyx_file.name]

            invalid_files = self._do_analyze(ref_valid_files + ref_invalid_files)

            self.assertEqual(ref_invalid_files, invalid_files)



class A():

    def __init__(self):
        self.var1 = 10
        self.attr = {}  # Break here

    def __dir__(self):
        return list(self.attr)


if __name__ == '__main__':
    a = A()
    print('TEST SUCEEDED!')



from robot.api.parsing import ModelTransformer, EmptyLine, Comment, CommentSection
from robotidy.decorators import check_start_end_line


class DiscardEmptySections(ModelTransformer):
    """
    Remove empty sections.
    Sections are considered empty if there are only empty lines inside.
    You can remove sections with only comments by setting ``allow_only_comments`` parameter to False:

        *** Variables ***
        # this section will be removed with``alow_only_comments`` parameter set to False

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/DiscardEmptySections.html for more examples.
    """

    def __init__(self, allow_only_comments: bool = True):
        # If False then sections with only with comments are considered to be empty
        self.allow_only_comments = allow_only_comments

    @check_start_end_line
    def visit_Section(self, node):  # noqa
        anything_but = (
            EmptyLine if self.allow_only_comments or isinstance(node, CommentSection) else (Comment, EmptyLine)
        )
        if all(isinstance(child, anything_but) for child in node.body):
            return None
        return node



__version__ = "0.0.1"
from typing import List

version_info: List[int] = [int(x) for x in __version__.split(".")]

import os
import sys


def get_src_folder() -> str:
    return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def import_robocorp_ls_core() -> None:
    """
    Helper function to make sure that robocorp_ls_core is imported properly
    (either in dev or in release mode).
    """

    try:
        import robocorp_ls_core
    except ImportError:
        log_contents = []
        use_folder = None
        try:
            src_folder = get_src_folder()
            log_contents.append("Source folder: %s" % (src_folder,))
            src_core_folder = os.path.abspath(
                os.path.join(src_folder, "..", "..", "robocorp-python-ls-core", "src")
            )

            if os.path.isdir(src_core_folder):
                log_contents.append("Dev mode detected. Found: %s" % (src_core_folder,))
                use_folder = src_core_folder

            else:
                # If not in dev, it's expected to be vendored just like it (so, it
                # should be alongside it).
                use_folder = src_folder
                log_contents.append("Using vendored mode. Found: %s" % (use_folder,))
                assert os.path.isdir(
                    use_folder
                ), "Expected: %s to exist and be a directory." % (use_folder,)

            sys.path.append(use_folder)
            import robocorp_ls_core
        except:
            try:
                if use_folder:
                    log_contents.append(
                        "%s contents:\n%s" % (use_folder, os.listdir(use_folder))
                    )
            except:
                log_contents.append("Error in os.listdir('%s')." % (use_folder,))
            raise ImportError(
                "Error importing robocorp_ls_core. Log: %s" % "\n".join(log_contents)
            )



"""
Spacing checkers
"""
import re
from collections import Counter

from robot.api import Token
from robot.parsing.model.blocks import Keyword, TestCase
from robot.parsing.model.statements import Comment, EmptyLine
from robot.parsing.model.visitor import ModelVisitor

from robocop.checkers import RawFileChecker, VisitorChecker
from robocop.rules import Rule, RuleParam, RuleSeverity
from robocop.utils import token_col, get_section_name, get_errors
from robocop.utils.misc import ROBOT_VERSION

rules = {
    "1001": Rule(
        rule_id="1001",
        name="trailing-whitespace",
        msg="Trailing whitespace at the end of line",
        severity=RuleSeverity.WARNING,
    ),
    "1002": Rule(
        rule_id="1002",
        name="missing-trailing-blank-line",
        msg="Missing trailing blank line at the end of file",
        severity=RuleSeverity.WARNING,
    ),
    "1003": Rule(
        RuleParam(
            name="empty_lines",
            default=2,
            converter=int,
            desc="number of empty lines required between sections",
        ),
        rule_id="1003",
        name="empty-lines-between-sections",
        msg="Invalid number of empty lines between sections ({{ empty_lines }}/{{ allowed_empty_lines }})",
        severity=RuleSeverity.WARNING,
    ),
    "1004": Rule(
        RuleParam(
            name="empty_lines",
            default=1,
            converter=int,
            desc="number of empty lines required between test cases",
        ),
        rule_id="1004",
        name="empty-lines-between-test-cases",
        msg="Invalid number of empty lines between test cases ({{ empty_lines }}/{{ allowed_empty_lines }})",
        severity=RuleSeverity.WARNING,
    ),
    "1005": Rule(
        RuleParam(
            name="empty_lines",
            default=1,
            converter=int,
            desc="number of empty lines required between keywords",
        ),
        rule_id="1005",
        name="empty-lines-between-keywords",
        msg="Invalid number of empty lines between keywords ({{ empty_lines }}/{{ allowed_empty_lines }})",
        severity=RuleSeverity.WARNING,
    ),
    "1006": Rule(
        rule_id="1006",
        name="mixed-tabs-and-spaces",
        msg="Inconsistent use of tabs and spaces in file",
        severity=RuleSeverity.WARNING,
    ),
    "1007": Rule(
        rule_id="1007",
        name="uneven-indent",
        msg="Line is {{ over_or_under }}-indented",
        severity=RuleSeverity.WARNING,
        docs="""
        Reported when line does not follow indent from the current block. 
        Example of rule violation::
        
            Keyword With Over Indented Setting
                [Documentation]  this is doc
                 [Arguments]  ${arg}  # over-indented
                   No Operation  # over-indented
                Pass
                No Operation
                Fail
        
        """,
    ),
    "1008": Rule(
        rule_id="1008",
        name="bad-indent",
        msg="Indent expected",
        severity=RuleSeverity.ERROR,
        docs="""
        Expecting indent. Example of rule violation::
        
             FOR  ${elem}  IN  ${list}
            Log  stuff  # content of FOR blocks should use bigger indentation than FOR header
             END
        
        """,
    ),
    "1009": Rule(
        RuleParam(
            name="empty_lines",
            default=0,
            converter=int,
            desc="number of empty lines allowed after section header",
        ),
        rule_id="1009",
        name="empty-line-after-section",
        msg="Too many empty lines after '{{ section_name }}' section header "
        "({{ empty_lines }}/{{ allowed_empty_lines }})",
        severity=RuleSeverity.WARNING,
        docs="""
        Empty lines after the section header are not allowed by default. Example of rule violation::
        
             *** Test Cases ***
             
             Resource  file.resource
        
        It can be configured using `empty_lines` parameter.
        """,
    ),
    "1010": Rule(
        rule_id="1010",
        name="too-many-trailing-blank-lines",
        msg="Too many blank lines at the end of file",
        severity=RuleSeverity.WARNING,
        docs="""There should be exactly one blank line at the end of the file""",
    ),
    "1011": Rule(
        rule_id="1011",
        name="misaligned-continuation",
        msg="Continuation marker should be aligned with starting row",
        severity=RuleSeverity.WARNING,
        docs="""
        Example of rule violation::
        
                Default Tags       default tag 1    default tag 2    default tag 3
            ...                default tag 4    default tag 5
            
                *** Test Cases ***
                Example
                    Do X    first argument    second argument    third argument
                  ...    fourth argument    fifth argument    sixth argument
        
        """,
    ),
    "1012": Rule(
        RuleParam(
            name="empty_lines",
            default=1,
            converter=int,
            desc="number of allowed consecutive empty lines",
        ),
        rule_id="1012",
        name="consecutive-empty-lines",
        msg="Too many consecutive empty lines ({{ empty_lines }}/{{ allowed_empty_lines }})",
        severity=RuleSeverity.WARNING,
        docs="""
        Example of rule violation::
        
            Keyword
                Step 1
                
                
                Step 2

        """,
    ),
    "1013": Rule(
        rule_id="1013",
        name="empty-lines-in-statement",
        msg="Multi-line statement with empty lines",
        severity=RuleSeverity.WARNING,
        docs="""
        Example of rule violation::
        
             Keyword
             ...  1
             # empty line in-between multiline statement
             ...  2
        
        """,
    ),
    "1014": Rule(
        rule_id="1014",
        name="variable-should-be-left-aligned",
        msg="Variable in Variable section should be left aligned",
        severity=RuleSeverity.ERROR,
        version=">=4.0",
        docs="""
        Example of rule violation::
        
            *** Variables ***
             ${VAR}  1
              ${VAR2}  2
        
        """,
    ),
    "1015": Rule(
        rule_id="1015",
        name="misaligned-continuation-row",
        msg="Each next continuation line should be aligned with the previous one",
        severity=RuleSeverity.WARNING,
        docs="""
        Example of rule violation::
        
            *** Settings ***
            Documentation      Here we have documentation for this suite.
            ...                Documentation is often quite long.
            ...
            ...            It can also contain multiple paragraphs.  # misaligned
            
            *** Test Cases ***
            Test
            [Tags]    you    probably    do    not    have    this    many
            ...      tags    in    real    life  # misaligned
        
        """,
    ),
    "1016": Rule(
        rule_id="1016",
        name="suite-setting-should-be-left-aligned",
        msg="Setting in Settings section should be left aligned",
        severity=RuleSeverity.ERROR,
        version=">=4.0",
        docs="""
        Example of rule violation::
        
            *** Settings ***
                Library  Collections
                Resource  data.resource
                Variables  vars.robot
        
        """,
    ),
}


class InvalidSpacingChecker(RawFileChecker):
    """Checker for trailing spaces and lines."""

    reports = (
        "trailing-whitespace",
        "missing-trailing-blank-line",
        "too-many-trailing-blank-lines",
    )

    def __init__(self):
        self.raw_lines = []
        super().__init__()

    def parse_file(self):
        self.raw_lines = []
        super().parse_file()
        if self.raw_lines:
            last_line = self.raw_lines[-1]
            if last_line in ["\n", "\r", "\r\n"]:
                self.report("too-many-trailing-blank-lines", lineno=len(self.raw_lines) + 1)
                return
            empty_lines = 0
            for line in self.raw_lines[::-1]:
                if not line.strip():
                    empty_lines += 1
                else:
                    break
                if empty_lines > 1:
                    self.report("too-many-trailing-blank-lines", lineno=len(self.raw_lines))
                    return
            if not empty_lines and not last_line.endswith(("\n", "\r")):
                self.report("missing-trailing-blank-line", lineno=len(self.raw_lines))

    def check_line(self, line, lineno):
        self.raw_lines.append(line)

        stripped_line = line.rstrip("\n\r")
        if stripped_line and stripped_line[-1] in [" ", "\t"]:
            self.report("trailing-whitespace", lineno=lineno, col=len(stripped_line))


class EmptyLinesChecker(VisitorChecker):
    """Checker for invalid spacing."""

    reports = (
        "empty-lines-between-sections",
        "empty-lines-between-test-cases",
        "empty-lines-between-keywords",
        "empty-line-after-section",
        "consecutive-empty-lines",
        "empty-lines-in-statement",
    )

    def verify_empty_lines(self, node, check_leading=True):
        """Verify number of consecutive empty lines inside node. Return number of trailing empty lines."""
        empty_lines = 0
        prev_node = None
        non_empty = check_leading  # if check_leading is set to False, we ignore leading empty lines
        for child in node.body:
            if isinstance(child, EmptyLine):
                if not non_empty:
                    continue
                empty_lines += 1
                prev_node = child
            else:
                non_empty = True
                if empty_lines > self.param("consecutive-empty-lines", "empty_lines"):
                    self.report(
                        "consecutive-empty-lines",
                        empty_lines=empty_lines,
                        allowed_empty_lines=self.param("consecutive-empty-lines", "empty_lines"),
                        node=prev_node,
                    )
                empty_lines = 0
        return empty_lines

    def visit_Statement(self, node):  # noqa
        prev_token = None
        for token in node.tokens:
            if token.type == Token.EOL:
                if prev_token:
                    self.report("empty-lines-in-statement", node=token)
                prev_token = token
            else:
                prev_token = None

    def visit_VariableSection(self, node):  # noqa
        self.verify_empty_lines(node, check_leading=False)
        self.generic_visit(node)

    def visit_SettingSection(self, node):  # noqa
        self.verify_empty_lines(node, check_leading=False)
        self.generic_visit(node)

    def verify_empty_lines_between_nodes(self, node, node_type, issue_name, allowed_empty_lines):
        last_index = len(node.body) - 1
        for index, child in enumerate(node.body):
            if not isinstance(child, node_type):
                continue
            empty_lines = self.verify_empty_lines(child)
            if allowed_empty_lines not in (empty_lines, -1) and index < last_index:
                self.report(
                    issue_name,
                    empty_lines=empty_lines,
                    allowed_empty_lines=allowed_empty_lines,
                    lineno=child.end_lineno,
                )
        self.generic_visit(node)

    def visit_TestCaseSection(self, node):  # noqa
        allowed_lines = -1 if self.templated_suite else self.param("empty-lines-between-test-cases", "empty_lines")
        self.verify_empty_lines_between_nodes(node, TestCase, "empty-lines-between-test-cases", allowed_lines)

    def visit_KeywordSection(self, node):  # noqa
        self.verify_empty_lines_between_nodes(
            node,
            Keyword,
            "empty-lines-between-keywords",
            self.param("empty-lines-between-keywords", "empty_lines"),
        )

    def visit_File(self, node):  # noqa
        for section in node.sections:
            self.check_empty_lines_after_section(section)
        for section in node.sections[:-1]:
            if not section.header:  # for comment section
                continue
            empty_lines = 0
            for child in reversed(section.body):
                if isinstance(child, (Keyword, TestCase)):
                    for statement in reversed(child.body):
                        if isinstance(statement, EmptyLine):
                            empty_lines += 1
                        else:
                            break
                if isinstance(child, EmptyLine):
                    empty_lines += 1
                elif isinstance(child, Comment):
                    continue
                else:
                    break
            if empty_lines != self.param("empty-lines-between-sections", "empty_lines"):
                self.report(
                    "empty-lines-between-sections",
                    empty_lines=empty_lines,
                    allowed_empty_lines=self.param("empty-lines-between-sections", "empty_lines"),
                    lineno=section.end_lineno,
                )
        super().visit_File(node)

    def check_empty_lines_after_section(self, section):
        empty_lines = []
        for child in section.body:
            if not isinstance(child, EmptyLine):
                break
            empty_lines.append(child)
        else:
            return
        if len(empty_lines) > self.param("empty-line-after-section", "empty_lines"):
            self.report(
                "empty-line-after-section",
                section_name=get_section_name(section),
                empty_lines=len(empty_lines),
                allowed_empty_lines=self.param("empty-line-after-section", "empty_lines"),
                node=empty_lines[-1],
            )


class InconsistentUseOfTabsAndSpacesChecker(VisitorChecker, ModelVisitor):
    """Checker for inconsistent use of tabs and spaces."""

    reports = ("mixed-tabs-and-spaces",)

    def __init__(self):
        self.found, self.tabs, self.spaces = False, False, False
        super().__init__()

    def visit_File(self, node):  # noqa
        self.found, self.tabs, self.spaces = False, False, False
        super().visit_File(node)

    def visit_Statement(self, node):  # noqa
        if self.found:
            return
        for token in node.get_tokens(Token.SEPARATOR):
            self.tabs = "\t" in token.value or self.tabs
            self.spaces = " " in token.value or self.spaces

            if self.tabs and self.spaces:
                self.report("mixed-tabs-and-spaces", node=node, lineno=1)
                self.found = True
                break


class UnevenIndentChecker(VisitorChecker):
    """Checker for indentation violations."""

    reports = (
        "uneven-indent",
        "bad-indent",
    )

    HEADERS = {
        Token.ARGUMENTS,
        Token.DOCUMENTATION,
        Token.SETUP,
        Token.TIMEOUT,
        Token.TEARDOWN,
        Token.TEMPLATE,
        Token.TAGS,
    }

    def visit_TestCaseSection(self, node):  # noqa
        self.check_standalone_comments_indent(node)

    def visit_KeywordSection(self, node):  # noqa
        self.check_standalone_comments_indent(node)

    def check_standalone_comments_indent(self, node):
        # comments before first test case / keyword
        for child in node.body:
            if (
                getattr(child, "type", "") == Token.COMMENT
                and getattr(child, "tokens", None)
                and child.tokens[0].type == Token.SEPARATOR
            ):
                self.report(
                    "uneven-indent",
                    over_or_under="over",
                    node=child,
                    col=token_col(child, Token.COMMENT),
                )
        self.generic_visit(node)

    def visit_TestCase(self, node):  # noqa
        self.check_indents(node)
        self.generic_visit(node)

    def visit_Keyword(self, node):  # noqa
        if not node.name.lstrip().startswith("#"):
            self.check_indents(node)
        self.generic_visit(node)

    def visit_ForLoop(self, node):  # noqa
        column_index = 2 if node.end is not None else 0
        self.check_indents(node, node.header.tokens[1].col_offset + 1, column_index)

    visit_For = visit_If = visit_While = visit_ForLoop

    def visit_Try(self, node):  # noqa
        column_index = 2 if node.end is not None else 0
        self.check_indents(node, node.header.tokens[1].col_offset + 1, column_index)
        while node.next:
            node = node.next
            self.check_indents(node, node.header.tokens[1].col_offset + 1, column_index)

    @staticmethod
    def get_indent(node):
        tokens = node.tokens if hasattr(node, "tokens") else node.header.tokens
        indent_len = 0
        for token in tokens:
            if token.type != Token.SEPARATOR:
                break
            indent_len += len(token.value.expandtabs(4))
        return indent_len

    def check_indents(self, node, req_indent=0, column_index=0, previous_indent=None):
        indents = []
        header_indents = []
        templated = self.is_templated(node)
        end_of_block = self.find_block_end(node) if not column_index else len(node.body)
        for index, child in enumerate(node.body):
            if index == end_of_block:
                break
            if child.lineno == node.lineno or isinstance(child, EmptyLine):
                continue
            indent_len = self.get_indent(child)
            if indent_len is None:
                continue
            if hasattr(child, "type") and child.type in UnevenIndentChecker.HEADERS:
                if templated:
                    header_indents.append((indent_len, child))
                else:
                    indents.append((indent_len, child))
            else:
                if indent_len < req_indent:
                    self.report("bad-indent", node=child)
                else:
                    indents.append((indent_len, child))
        if not column_index:
            self.validate_standalone_comments(node.body[end_of_block:])
        previous_indent = self.validate_indent_lists(indents, previous_indent)
        if templated:
            self.validate_indent_lists(header_indents)
        if getattr(node, "orelse", None):
            self.check_indents(node.orelse, req_indent, column_index, previous_indent)

    def validate_indent_lists(self, indents, previous_indent=None):
        counter = Counter(indent[0] for indent in indents)
        if previous_indent:  # in case of continuing blocks, ie if else blocks, remember the indent
            counter = previous_indent + counter
        elif len(indents) < 2:
            return counter
        if len(counter) == 1:  # everything have the same indent
            return counter
        common_indent = counter.most_common(1)[0][0]
        for indent in indents:
            if indent[0] != common_indent:
                self.report(
                    "uneven-indent",
                    over_or_under="over" if indent[0] > common_indent else "under",
                    node=indent[1],
                    col=indent[0] + 1,
                )
        return counter

    def validate_standalone_comments(self, comments_and_eols):
        """
        Report any comment that does not start from col 1.

        :param comments_and_eols: list of comments and empty lines (outside keyword and test case definitions)
        """
        for child in comments_and_eols:
            if getattr(child, "type", "invalid") != Token.COMMENT:
                continue
            col = token_col(child, Token.COMMENT)
            if col != 1:
                self.report("uneven-indent", over_or_under="over", node=child, col=col)

    @staticmethod
    def find_block_end(node):
        """
        Find where the keyword/test case/block ends. If there are only comments and new lines left, the
        first comment that starts from col 1 is considered outside your block::

            Keyword
                Line
                # comment
                Other Line
               # comment belonging to Keyword

            # This should not belong to Keyword
              # Since there was comment starting from col 1, this comment is also outside block
        """
        for index, child in reversed(list(enumerate(node.body))):
            node_type = getattr(child, "type", "")
            if not node_type:
                return len(node.body) - 1
            if node_type not in (Token.COMMENT, Token.EOL):
                break
        else:
            return len(node.body) - 1
        for block_index, child in enumerate(node.body[index + 1 :]):
            if getattr(child, "type", "invalid") == Token.COMMENT and token_col(child, Token.COMMENT) == 1:
                return block_index + index + 1
        return len(node.body) - 1

    @staticmethod
    def is_templated(node):
        if not isinstance(node, TestCase):
            return False
        for child in node.body:
            if hasattr(child, "type") and child.type == "TEMPLATE":
                return True
        return False


class MisalignedContinuation(VisitorChecker, ModelVisitor):
    """Checker for misaligned continuation line markers."""

    reports = (
        "misaligned-continuation",
        "misaligned-continuation-row",
    )

    def visit_Statement(self, node):  # noqa
        if not node.data_tokens:
            return
        starting_row = self.get_indent(node.tokens)
        first_column, indent = 0, 0
        for index, line in enumerate(node.lines):
            if index == 0:
                starting_row = self.get_indent(line)
                if node.type == Token.TAGS:
                    first_column = self.first_line_indent(line, node.type, Token.ARGUMENT)
                continue
            indent = 0
            for token in line:
                if token.type == Token.SEPARATOR:  # count possible indent before or after ...
                    indent += len(token.value.expandtabs(4))
                elif token.type == Token.CONTINUATION:
                    if indent != starting_row:
                        self.report(
                            "misaligned-continuation",
                            lineno=token.lineno,
                            col=token.col_offset + 1,
                        )
                        break
                    indent = 0
                elif token.type != Token.EOL and token.value.strip():  # ignore trailing whitespace
                    if first_column:
                        if indent != first_column:
                            self.report(
                                "misaligned-continuation-row",
                                node=token,
                                col=token.col_offset + 1,
                            )
                    else:
                        first_column = indent
                    break  # check only first value

    @staticmethod
    def get_indent(tokens):
        indent_len = 0
        for token in tokens:
            if token.type != Token.SEPARATOR:
                break
            indent_len += len(token.value.expandtabs(4))
        return indent_len

    @staticmethod
    def first_line_indent(tokens, from_tok, search_for):
        """
        Find indent required for other lines to match indentation of first line.

        [from_token]     <search_for>
        ...<-   pos   ->

        :param tokens: statement first line tokens
        :param from_tok: start counting separator after finding from_tok token
        :param search_for: stop counting after finding search_for token
        :return: pos: length of indent
        """
        pos = 0
        found = False
        for token in tokens:
            if not found:
                if token.type == from_tok:
                    found = True
                    # subtract 3 to adjust for ... length in 2nd line
                    pos += len(token.value) - 3
            elif token.type == Token.SEPARATOR:
                pos += len(token.value.expandtabs(4))
            elif token.type == search_for:
                return pos
        return 0  # 0 will ignore first line indent and compare to 2nd line only


class LeftAlignedChecker(VisitorChecker):
    """Checker for left align."""

    reports = (
        "variable-should-be-left-aligned",
        "suite-setting-should-be-left-aligned",
    )

    suite_settings = {
        "documentation": "Documentation",
        "suitesetup": "Suite Setup",
        "suiteteardown": "Suite Teardown",
        "metadata": "Metadata",
        "testsetup": "Test Setup",
        "testteardown": "Test Teardown",
        "testtemplate": "Test Template",
        "testtimeout": "Test Timeout",
        "forcetags": "Force Tags",
        "defaulttags": "Default Tags",
        "library": "Library",
        "resource": "Resource",
        "variables": "Variables",
    }

    def visit_VariableSection(self, node):  # noqa
        for child in node.body:
            if not child.data_tokens:
                continue
            token = child.data_tokens[0]
            if token.type == Token.VARIABLE and (token.value == "" or token.value.startswith(" ")):
                if token.value or not child.get_token(Token.ARGUMENT):
                    pos = len(token.value) - len(token.value.lstrip()) + 1
                else:
                    pos = child.get_token(Token.ARGUMENT).col_offset + 1
                self.report("variable-should-be-left-aligned", lineno=token.lineno, col=pos)

    def visit_SettingSection(self, node):  # noqa
        for child in node.body:
            for error in get_errors(child):
                if "Non-existing setting" in error:
                    self.parse_error(child, error)

    def parse_error(self, node, error):
        setting_error = re.search("Non-existing setting '(.*)'.", error)
        if not setting_error:
            return
        setting_error = setting_error.group(1)
        if not setting_error:
            setting_cand = node.get_token(Token.COMMENT)
            if setting_cand and setting_cand.value.replace(" ", "").lower() in self.suite_settings:
                self.report(
                    "suite-setting-should-be-left-aligned",
                    node=setting_cand,
                    col=setting_cand.col_offset + 1,
                )
        elif not setting_error[0].strip():  # starts with space/tab
            suite_sett_cand = setting_error.replace(" ", "").lower()
            for setting in self.suite_settings:
                if suite_sett_cand.startswith(setting):
                    indent = len(setting_error) - len(setting_error.lstrip())
                    self.report(
                        "suite-setting-should-be-left-aligned",
                        node=node,
                        col=indent + 1,
                    )
                    break



def m1():
    _a = 'm1'  # break 1 here


def m2():  # @DontTrace
    m1()
    _a = 'm2'


def m3():
    m2()
    _a = 'm3'  # break 2 here


if __name__ == '__main__':
    m3()

    print('TEST SUCEEDED')



import sys
import threading
import pytest
from tests_python.debugger_unittest import IS_PY36_OR_GREATER, IS_CPYTHON
from tests_python.debug_constants import TEST_CYTHON

pytestmark = pytest.mark.skipif(not IS_PY36_OR_GREATER or not IS_CPYTHON or not TEST_CYTHON, reason='Requires CPython >= 3.6')


def get_foo_frame():
    frame = sys._getframe()
    return frame


class CheckClass(object):

    def collect_info(self):
        from _pydevd_frame_eval import pydevd_frame_evaluator
        thread_info = pydevd_frame_evaluator.get_thread_info_py()
        self.thread_info = thread_info


@pytest.mark.parametrize('_times', range(2))
def test_thread_info(_times):
    obj = CheckClass()
    obj.collect_info()
    assert obj.thread_info.additional_info is not None
    assert not obj.thread_info.is_pydevd_thread
    thread_info = obj.thread_info
    obj.collect_info()
    assert obj.thread_info is thread_info

    obj = CheckClass()
    t = threading.Thread(target=obj.collect_info)
    t.is_pydev_daemon_thread = True
    t.start()
    t.join()

    assert obj.thread_info.additional_info is None
    assert obj.thread_info.is_pydevd_thread


def method():
    return sys._getframe()


@pytest.fixture
def _custom_global_dbg():
    from _pydevd_bundle.pydevd_constants import GlobalDebuggerHolder
    from pydevd import PyDB
    curr = GlobalDebuggerHolder.global_dbg
    PyDB()  # Will make itself current
    yield
    GlobalDebuggerHolder.global_dbg = curr


@pytest.mark.parametrize('_times', range(2))
def test_func_code_info(_times, _custom_global_dbg):
    from _pydevd_frame_eval import pydevd_frame_evaluator
    # Must be called before get_func_code_info_py to initialize the _code_extra_index.
    thread_info = pydevd_frame_evaluator.get_thread_info_py()

    func_info = pydevd_frame_evaluator.get_func_code_info_py(thread_info, method(), method.__code__)
    assert func_info.co_filename is method.__code__.co_filename
    func_info2 = pydevd_frame_evaluator.get_func_code_info_py(thread_info, method(), method.__code__)
    assert func_info is func_info2

    some_func = eval('lambda:sys._getframe()')
    func_info3 = pydevd_frame_evaluator.get_func_code_info_py(thread_info, some_func(), some_func.__code__)
    del some_func
    del func_info3

    some_func = eval('lambda:sys._getframe()')
    pydevd_frame_evaluator.get_func_code_info_py(thread_info, some_func(), some_func.__code__)
    func_info = pydevd_frame_evaluator.get_func_code_info_py(thread_info, some_func(), some_func.__code__)
    assert pydevd_frame_evaluator.get_func_code_info_py(thread_info, some_func(), some_func.__code__) is func_info


def test_generate_code_with_breakpoints():
    from _pydevd_frame_eval.pydevd_frame_evaluator import generate_code_with_breakpoints_py
    from _pydevd_frame_eval.pydevd_frame_evaluator import get_cached_code_obj_info_py

    def create_breakpoints_dict(lines):
        return dict((line, None) for line in lines)

    def method():
        a = 1
        a = 2
        a = 3

    breakpoint_found, new_code = generate_code_with_breakpoints_py(
        method.__code__,
        create_breakpoints_dict([method.__code__.co_firstlineno + 1, method.__code__.co_firstlineno + 2])
    )

    assert breakpoint_found
    with pytest.raises(AssertionError):
        # We must use the cached one directly (in the real-world, this would indicate a reuse
        # of the code object -- which is related to generator handling).
        generate_code_with_breakpoints_py(
            new_code,
            create_breakpoints_dict([method.__code__.co_firstlineno + 1])
        )

    cached_value = get_cached_code_obj_info_py(new_code)
    breakpoint_found, force_stay_in_untraced_mode = cached_value.compute_force_stay_in_untraced_mode(
        create_breakpoints_dict([method.__code__.co_firstlineno + 1]))

    assert breakpoint_found
    assert force_stay_in_untraced_mode

    # i.e.: no breakpoints match (stay in untraced mode)
    breakpoint_found, force_stay_in_untraced_mode = cached_value.compute_force_stay_in_untraced_mode(
        create_breakpoints_dict([method.__code__.co_firstlineno + 10]))

    assert not breakpoint_found
    assert force_stay_in_untraced_mode

    # i.e.: one of the breakpoints match (stay in untraced mode)
    breakpoint_found, force_stay_in_untraced_mode = cached_value.compute_force_stay_in_untraced_mode(
        create_breakpoints_dict([method.__code__.co_firstlineno + 2]))

    assert breakpoint_found
    assert force_stay_in_untraced_mode

    # i.e.: one of the breakpoints doesn't match (leave untraced mode)
    breakpoint_found, force_stay_in_untraced_mode = cached_value.compute_force_stay_in_untraced_mode(
        create_breakpoints_dict([method.__code__.co_firstlineno + 3]))

    assert breakpoint_found
    assert not force_stay_in_untraced_mode




class TestProperty(object):
    def __init__(self, name = "Default"):
        self._x = None
        self.name = name

    def get_name(self):
        return self.__name


    def set_name(self, value):
        self.__name = value


    def del_name(self):
        del self.__name
    name = property(get_name, set_name, del_name, "name's docstring")

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        self._x = value

    @x.deleter
    def x(self):
        del self._x

def main():
    """
    """
    testObj = TestProperty()
    testObj.x = 10
    val = testObj.x
    
    testObj.name = "Pydev"
    debugType = testObj.name
    print('TEST SUCEEDED!')
    
if __name__ == '__main__':
    main()


import os.path
from pathlib import Path
import sys
from typing import List, Dict

from robocorp_ls_core.robotframework_log import get_logger


log = get_logger(__name__)


def _import_robocop():
    try:
        import robocop
    except ImportError:
        _parent_dir = os.path.dirname(__file__)
        _robocop_dir = os.path.join(_parent_dir, "libs", "robocop_lib")
        if not os.path.exists(_robocop_dir):
            raise RuntimeError("Expected: %s to exist." % (_robocop_dir,))
        sys.path.append(_robocop_dir)
        import robocop  # @UnusedImport

        log.info("Using vendored Robocop")

    log.info("Robocop module: %s", robocop)


def collect_robocop_diagnostics(
    project_root: Path, ast_model, filename: str, source: str
) -> List[Dict]:

    _import_robocop()

    import robocop
    from robocop.config import Config
    from robocop.utils import issues_to_lsp_diagnostic

    filename_parent = Path(filename).parent
    if filename_parent.exists():
        config = Config(root=filename_parent)
    else:
        # Unsaved files.
        config = Config(root=project_root)
    robocop_runner = robocop.Robocop(config=config)
    robocop_runner.reload_config()

    issues = robocop_runner.run_check(ast_model, filename, source)
    diag_issues = issues_to_lsp_diagnostic(issues)
    return diag_issues



"""
To use, create a SystemMutex, check if it was acquired (get_mutex_aquired()) and if acquired the
mutex is kept until the instance is collected or release_mutex is called.

I.e.:

    mutex = SystemMutex('my_unique_name')
    if mutex.get_mutex_aquired():
        print('acquired')
    else:
        print('not acquired')
    
    
Or to keep trying to get the mutex until a given timeout elapses:

    with timed_acquire_mutex('mutex_name'):
        # Do something without any racing condition with other processes
        ...

License: Dual-licensed under LGPL and Apache 2.0

Copyright: Brainwy Software
Author: Fabio Zadrozny
"""

import re
import sys
import tempfile
import time
import traceback
import weakref

from robocorp_ls_core.constants import NULL
from robocorp_ls_core.robotframework_log import get_logger
import threading
from typing import ContextManager


log = get_logger(__name__)


def check_valid_mutex_name(mutex_name):
    # To be windows/linux compatible we can't use non-valid filesystem names
    # (as on linux it's a file-based lock).

    regexp = re.compile(r'[\*\?"<>|/\\:]')
    result = regexp.findall(mutex_name)
    if result is not None and len(result) > 0:
        raise AssertionError("Mutex name is invalid: %s" % (mutex_name,))


_mutex_name_to_info: "weakref.WeakValueDictionary[str, SystemMutex]" = (
    weakref.WeakValueDictionary()
)
_lock = threading.Lock()


def get_tid():
    return threading.get_ident()  # @UndefinedVariable


def _mark_prev_acquired_in_thread(system_mutex):
    with _lock:
        _mutex_name_to_info[system_mutex.mutex_name] = system_mutex


def _verify_prev_acquired_in_thread(mutex_name):
    with _lock:
        system_mutex = _mutex_name_to_info.get(mutex_name)
        if (
            system_mutex is not None
            and system_mutex.get_mutex_aquired()
            and not system_mutex.disposed
            and system_mutex.thread_id == get_tid()
        ):
            raise RuntimeError(
                "Error: this thread has already acquired a SystemMutex(%s) and it's not a reentrant mutex (so, this would never work)!"
                % (mutex_name,)
            )


if sys.platform == "win32":

    import os

    class SystemMutex(object):
        def __init__(
            self, mutex_name, check_reentrant=True, log_info=False, base_dir=None
        ):
            """
            :param check_reentrant:
                Should only be False if this mutex is expected to be released in
                a different thread.
            """
            if base_dir is None:
                base_dir = tempfile.gettempdir()
            check_valid_mutex_name(mutex_name)
            self.mutex_name = mutex_name
            self.thread_id = get_tid()
            filename = os.path.join(base_dir, mutex_name)
            try:
                os.unlink(filename)
            except Exception:
                pass
            try:
                handle = os.open(filename, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                try:
                    try:
                        pid = str(os.getpid())
                    except Exception:
                        pid = "unable to get pid"
                    os.write(handle, pid)
                except Exception:
                    pass  # Ignore this as it's pretty much optional
            except Exception:
                self._release_mutex = NULL
                self._acquired = False
                if check_reentrant:
                    _verify_prev_acquired_in_thread(mutex_name)
            else:

                def release_mutex(*args, **kwargs):
                    # Note: can't use self here!
                    if not getattr(release_mutex, "called", False):
                        release_mutex.called = True
                        try:
                            os.close(handle)
                        except Exception:
                            traceback.print_exc()
                        try:
                            # Removing is optional as we'll try to remove on startup anyways (but
                            # let's do it to keep the filesystem cleaner).
                            os.unlink(filename)
                        except Exception:
                            pass

                # Don't use __del__: this approach doesn't have as many pitfalls.
                self._ref = weakref.ref(self, release_mutex)

                self._release_mutex = release_mutex
                self._acquired = True
                _mark_prev_acquired_in_thread(self)

        @property
        def disposed(self):
            if not self._acquired:
                return True
            release_mutex = self._release_mutex
            if release_mutex is NULL:
                return True
            if getattr(release_mutex, "called", False):
                return True
            return False

        def get_mutex_aquired(self):
            return self._acquired

        def release_mutex(self):
            self._release_mutex()


else:  # Linux
    import os
    import fcntl  # @UnresolvedImport

    class SystemMutex(object):
        def __init__(
            self, mutex_name, check_reentrant=True, log_info=False, base_dir=None
        ):
            """
            :param check_reentrant:
                Should only be False if this mutex is expected to be released in
                a different thread.
            """
            if base_dir is None:
                base_dir = tempfile.gettempdir()
            check_valid_mutex_name(mutex_name)
            self.mutex_name = mutex_name
            self.thread_id = get_tid()
            filename = os.path.join(base_dir, mutex_name)
            try:
                handle = open(filename, "a+")
                fcntl.flock(handle, fcntl.LOCK_EX | fcntl.LOCK_NB)
                handle.write(str(os.getpid()) + "\n")
                handle.flush()
            except Exception:
                self._release_mutex = NULL
                self._acquired = False
                if check_reentrant:
                    _verify_prev_acquired_in_thread(mutex_name)
                try:
                    handle.close()
                except Exception:
                    pass

                # It was unable to get the lock
                if log_info:
                    try:
                        try:
                            with open(filename, "r") as stream:
                                curr_pid = stream.read().strip().split("\n")[-1]
                        except:
                            log.exception("Unable to get locking pid.")
                            curr_pid = "<unable to get locking pid>"

                        try:
                            from robocorp_ls_core.basic import is_process_alive

                            is_alive = is_process_alive(int(curr_pid))
                        except:
                            is_alive = "<unknown>"
                        log.info(
                            "Current pid holding lock: %s. Alive: %s. This pid: %s Filename: %s",
                            curr_pid,
                            is_alive,
                            os.getpid(),
                            filename,
                        )
                    except:
                        from robocorp_ls_core.robotframework_log import get_log_level

                        if get_log_level() >= 2:
                            log.exception("Error getting lock info on failure.")
            else:

                def release_mutex(*args, **kwargs):
                    # Note: can't use self here!
                    if not getattr(release_mutex, "called", False):
                        release_mutex.called = True
                        try:
                            fcntl.flock(handle, fcntl.LOCK_UN)
                        except Exception:
                            traceback.print_exc()
                        try:
                            handle.close()
                        except Exception:
                            traceback.print_exc()
                        try:
                            # Removing is pretty much optional (but let's do it to keep the
                            # filesystem cleaner).
                            os.unlink(filename)
                        except Exception:
                            pass

                # Don't use __del__: this approach doesn't have as many pitfalls.
                self._ref = weakref.ref(self, release_mutex)

                self._release_mutex = release_mutex
                self._acquired = True
                _mark_prev_acquired_in_thread(self)

        @property
        def disposed(self):
            if not self._acquired:
                return True
            release_mutex = self._release_mutex
            if release_mutex is NULL:
                return True
            if getattr(release_mutex, "called", False):
                return True
            return False

        def get_mutex_aquired(self):
            return self._acquired

        def release_mutex(self):
            self._release_mutex()


class _MutexHandle(object):
    def __init__(self, system_mutex, mutex_name):
        self._system_mutex = system_mutex
        self._mutex_name = mutex_name
        # log.info("Obtained mutex: %s in pid: %s", mutex_name, os.getpid())

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        self._system_mutex.release_mutex()
        # log.info("Released mutex: %s in pid: %s", self._mutex_name, os.getpid())


def timed_acquire_mutex(
    mutex_name, timeout=20, sleep_time=0.15, check_reentrant=True, base_dir=None
) -> ContextManager:
    """
    Acquires the mutex given its name, a number of attempts and a time to sleep between each attempt.

    :throws RuntimeError if it was not possible to get the mutex in the given time.

    To be used as:

    with timed_acquire_mutex('mutex_name'):
        # Do something without any racing condition with other processes
        ...


    :param check_reentrant:
        Should only be False if this mutex is expected to be released in
        a different thread.
    """
    finish_at = time.time() + timeout
    while True:
        last_attempt = time.time() >= finish_at
        mutex = SystemMutex(
            mutex_name,
            check_reentrant=check_reentrant,
            log_info=last_attempt,
            base_dir=base_dir,
        )
        if not mutex.get_mutex_aquired():
            if last_attempt:
                log.info("Unable to obtain mutex: %s", mutex_name)
                raise RuntimeError(
                    "Could not get mutex: %s after: %s secs." % (mutex_name, timeout)
                )

            time.sleep(sleep_time)

            del mutex
        else:
            return _MutexHandle(mutex, mutex_name)


def generate_mutex_name(target_name, prefix=""):
    """
    A mutex name must be a valid filesystem path, so, this generates a hash
    that can be used in case the original name would have conflicts.
    """
    import hashlib

    if not isinstance(target_name, bytes):
        target_name = target_name.encode("utf-8")

    return prefix + (hashlib.sha224(target_name).hexdigest()[:16])



class MyVersion(object):
    pass


class MyPackage(object):
    pass


class MyName(object):
    pass


__version__ = MyVersion()
__package__ = MyPackage()
__name__ = MyName()

print('TEST SUCEEDED')  # Break here



try:
    from django.conf.urls import url
except ImportError:
    from django.urls import re_path as url

from . import views

urlpatterns = [
    url(r'^$', views.index, name='index'),
    url(r'^name$', views.get_name, name='name'),
    url(r'^template_error2$', views.template_error2, name='template_error2'),
    url(r'^template_error$', views.template_error, name='template_error'),
    url(r'^inherits$', views.inherits, name='inherits'),
    url(r'^no_var_error$', views.no_var_error, name='no_var_error'),
]



import time
import sys

try:
    try:
        from PySide import QtCore  # @UnresolvedImport
    except:
        from PySide2 import QtCore  # @UnresolvedImport
except:
    try:
        from PyQt4 import QtCore
    except:
        from PyQt5 import QtCore

# Using a QRunnable
# http://doc.qt.nokia.com/latest/qthreadpool.html
# Note that a QRunnable isn't a subclass of QObject and therefore does
# not provide signals and slots.
class Runnable(QtCore.QRunnable):

    def run(self):
        count = 0
        app = QtCore.QCoreApplication.instance()
        while count < 5:
            print("Increasing")  # break here
            count += 1
        app.quit()


app = QtCore.QCoreApplication([])
runnable = Runnable()
QtCore.QThreadPool.globalInstance().start(runnable)
app.exec_()
QtCore.QThreadPool.globalInstance().waitForDone()
print('TEST SUCEEDED!')



from robocop import checkers
from robocop.config import Config
from robocop.run import Robocop, run_robocop
from robocop.version import __version__

__all__ = ["__version__", "run_robocop", "checkers", "Robocop", "Config"]



def generator():
    yield 1  # stop 1
    yield 2  # stop 4


def main():
    for i in generator():  # stop 3
        print(i)  # stop 2


if __name__ == '__main__':
    main()
    print('TEST SUCEEDED!')



import sys
from typing import (
    TypeVar,
    Any,
    Optional,
    List,
    Sequence,
    Tuple,
    Iterable,
    Generic,
    Iterator,
    Callable,
    Union,
    Hashable,
    Dict,
)
from robocorp_ls_core.protocols import (
    Sentinel,
    IMonitor,
    IDocument,
    IWorkspace,
    IConfig,
    IDocumentSelection,
    ITestInfoFromSymbolsCacheTypedDict,
)
from robocorp_ls_core.constants import NULL
from robocorp_ls_core.protocols import TypedDict
import enum
from robocorp_ls_core.lsp import (
    LocationTypedDict,
    RangeTypedDict,
    MarkupContentTypedDict,
    CompletionItemTypedDict,
)
import typing
from robocorp_ls_core.ordered_set import OrderedSet
from contextlib import contextmanager

if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass


else:
    from typing import Protocol

T = TypeVar("T")
Y = TypeVar("Y", covariant=True)

# We don't want to import robot in this case (just do it when type-checking).
class IRobotToken(Protocol):
    SETTING_HEADER: str
    VARIABLE_HEADER: str
    TESTCASE_HEADER: str
    KEYWORD_HEADER: str
    COMMENT_HEADER: str

    TESTCASE_NAME: str
    KEYWORD_NAME: str

    DOCUMENTATION: str
    SUITE_SETUP: str
    SUITE_TEARDOWN: str
    METADATA: str
    TEST_SETUP: str
    TEST_TEARDOWN: str
    TEST_TEMPLATE: str
    TEST_TIMEOUT: str
    FORCE_TAGS: str
    DEFAULT_TAGS: str
    LIBRARY: str
    RESOURCE: str
    VARIABLES: str
    SETUP: str
    TEARDOWN: str
    TEMPLATE: str
    TIMEOUT: str
    TAGS: str
    ARGUMENTS: str
    # Use ´RETURN_SETTING` type instead of `RETURN`. `[Return]` is deprecated and
    # `RETURN` type will be used with `RETURN` statement in the future.
    RETURN: str
    RETURN_SETTING: str

    NAME: str
    VARIABLE: str
    ARGUMENT: str
    ASSIGN: str
    KEYWORD: str
    WITH_NAME: str
    FOR: str
    FOR_SEPARATOR: str
    END: str
    IF: str
    INLINE_IF: str
    ELSE_IF: str
    ELSE: str
    TRY: str
    EXCEPT: str
    FINALLY: str
    AS: str
    WHILE: str
    RETURN_STATEMENT: str
    CONTINUE: str
    BREAK: str

    SEPARATOR: str
    COMMENT: str
    CONTINUATION: str
    EOL: str
    EOS: str

    ERROR: str
    FATAL_ERROR: str

    type: str
    value: str
    lineno: int
    col_offset: int
    error: Any

    @property
    def end_col_offset(self) -> int:
        pass

    def tokenize_variables(self) -> Iterator["IRobotToken"]:
        pass


class IRobotVariableMatch(Protocol):
    string: str
    identifier: str
    base: Optional[str]
    items: Tuple[str, ...]
    start: int
    end: int

    @property
    def name(self) -> str:
        pass

    @property
    def before(self) -> str:
        pass

    @property
    def match(self) -> str:
        pass

    @property
    def after(self) -> str:
        pass


class INode(Protocol):
    lineno: int
    end_lineno: int
    col_offset: int
    end_col_offset: int
    tokens: List[IRobotToken]

    def get_token(self, name: str) -> Optional[IRobotToken]:
        pass

    def get_tokens(self, name: str) -> List[IRobotToken]:
        pass


class ILibraryImportNode(INode, Protocol):
    name: str
    alias: Optional[str]


class IKeywordNode(INode, Protocol):
    name: str


class IResourceImportNode(INode, Protocol):
    name: str


class IVariableImportNode(INode, Protocol):
    name: str


class NodeInfo(Generic[Y]):
    stack: Tuple[INode, ...]
    node: Y

    __slots__ = ["stack", "node"]

    def __init__(self, stack, node):
        self.stack = stack
        self.node = node

    def __str__(self):
        return f"NodeInfo({self.node.__class__.__name__})"

    __repr__ = __str__


class TokenInfo:

    __slots__ = ["stack", "node", "token"]

    def __init__(self, stack: Tuple[INode, ...], node: INode, token: IRobotToken):
        self.stack = stack
        self.node = node
        self.token = token

    def __str__(self):
        return f"TokenInfo({self.token.value} -- in: {self.node.__class__.__name__})"

    __repr__ = __str__


class VarTokenInfo:

    __slots__ = ["stack", "node", "token", "var_identifier", "context"]

    CONTEXT_UNDEFINED = 0
    CONTEXT_EXPRESSION = 1

    def __init__(
        self,
        stack: Tuple[INode, ...],
        node: INode,
        token: IRobotToken,
        var_identifier: str,
        context: int = CONTEXT_UNDEFINED,
    ):
        """
        :param var_identifier: One of: $,@,%
        :param context:
        """
        self.stack = stack
        self.node = node
        self.token = token
        self.var_identifier = var_identifier
        self.context = context

    def __str__(self):
        return f"VarTokenInfo({self.token.value} -- id: {self.var_identifier} -- in: {self.node.__class__.__name__})"

    __repr__ = __str__


class KeywordUsageInfo:
    __slots__ = [
        "stack",
        "node",
        "token",
        "name",
        "_is_argument_usage",
    ]

    def __init__(
        self,
        stack: Tuple[INode, ...],
        node: INode,
        token: IRobotToken,
        name: str,
        is_argument_usage: bool = False,
    ):
        self.stack = stack
        self.node = node
        self.token = token
        self.name = name
        self._is_argument_usage = is_argument_usage

    def __repr__(self):
        if self._is_argument_usage:
            return f"KeywordUsageInfo({self.name} (argument usage))"
        else:
            return f"KeywordUsageInfo({self.name})"

    __str__ = __repr__


class IKeywordArg(Protocol):
    @property
    def original_arg(self) -> str:
        pass

    @property
    def arg_name(self) -> str:
        pass

    @property
    def is_keyword_arg(self) -> bool:
        pass

    @property
    def is_star_arg(self) -> bool:
        pass

    def is_arg_type_set(self) -> bool:
        pass

    @property
    def arg_type(self) -> Optional[str]:
        pass

    def is_default_value_set(self) -> bool:
        pass

    @property
    def default_value(self) -> Optional[str]:
        pass


class ILibraryDoc(Protocol):
    filename: str
    name: str
    source: str
    symbols_cache: Optional["ISymbolsCache"]
    inits: list
    doc_format: str
    keywords: List["IKeywordDoc"]


class IKeywordDoc(Protocol):
    name: str
    tags: Tuple[str, ...]
    lineno: int
    doc: str

    @property
    def args(self) -> Tuple[IKeywordArg, ...]:
        pass

    @property
    def libdoc(self) -> ILibraryDoc:
        pass

    @property
    def deprecated(self) -> bool:
        pass

    @property
    def source(self) -> str:
        pass

    @property
    def doc_format(self) -> str:
        pass

    def to_dictionary(self) -> dict:
        pass


class ILibraryDocOrError(Protocol):
    library_doc: Optional[ILibraryDoc]
    error: Optional[str]


class IRobotDocument(IDocument, Protocol):
    def get_type(self) -> str:
        pass

    def get_ast(self) -> Any:
        pass

    def get_python_ast(self) -> Optional[Any]:
        pass

    def get_yaml_contents(self) -> Optional[Any]:
        pass

    symbols_cache: Optional["ISymbolsCache"]


class ISymbolsJsonListEntry(TypedDict):
    name: str
    kind: int  # SymbolKind
    location: LocationTypedDict
    containerName: str


class ISymbolKeywordInfo(Protocol):
    name: str

    def get_documentation(self) -> MarkupContentTypedDict:
        """
        Note: It should be computed on demand (and can be slow).
        """


class ISymbolsCache(Protocol):
    def get_uri(self) -> Optional[str]:
        pass

    def has_keyword_usage(self, normalized_keyword_name: str) -> bool:
        pass

    def get_json_list(self) -> List[ISymbolsJsonListEntry]:
        pass

    def get_library_info(self) -> Optional[ILibraryDoc]:
        pass

    def get_doc(self) -> Optional[IRobotDocument]:
        pass

    def get_test_info(self) -> Optional[List[ITestInfoFromSymbolsCacheTypedDict]]:
        pass

    def iter_keyword_info(self) -> Iterator[ISymbolKeywordInfo]:
        pass


class ICompletionContextWorkspaceCaches(Protocol):
    cache_hits: int

    def on_file_changed(self, filename: str):
        pass

    def on_updated_document(self, uri: str, document: Optional[IRobotDocument]):
        pass

    def clear_caches(self):
        pass

    def dispose(self):
        pass

    def get_cached_dependency_graph(
        self, cache_key: Hashable
    ) -> Optional["ICompletionContextDependencyGraph"]:
        pass

    @contextmanager
    def invalidation_tracker(self):
        """
        Note that it's possible that changes happen in-flight. This means that
        we must track changes while the dependency graph is being calculated.

        So, one must do something as:

        with caches.invalidation_tracker() as invalidation_tracker:
            ... compute dependency graph
            caches.cache_dependency_graph(cache_key, dependency_graph, invalidation_tracker)
        """

    def cache_dependency_graph(
        self,
        cache_key: Hashable,
        dependency_graph: "ICompletionContextDependencyGraph",
        invalidation_tracker,
    ) -> None:
        pass


class IRobotWorkspace(IWorkspace, Protocol):
    completion_context_workspace_caches: ICompletionContextWorkspaceCaches
    libspec_manager: Any

    def iter_all_doc_uris_in_workspace(
        self, extensions: Tuple[str, ...]
    ) -> Iterable[str]:
        pass


class IKeywordFound(Protocol):
    """
    :ivar completion_context:
        This may be a new completion context, created when a new document is
        being analyzed (the keyword was created for that completion context).
        For libraries the initial completion context is passed.
    :ivar source:
        Source where the keyword was found.
    :ivar lineno:
        Line where it was found (0-based).
    """

    @property
    def keyword_name(self) -> str:
        pass

    @property
    def keyword_args(self) -> Sequence[IKeywordArg]:
        pass

    def is_deprecated(self) -> bool:
        pass

    def compute_docs_with_signature(self) -> MarkupContentTypedDict:
        pass

    def compute_docs_without_signature(self) -> MarkupContentTypedDict:
        pass

    completion_context: Optional["ICompletionContext"]

    completion_item_kind: int = -1

    @property
    def source(self) -> str:
        """
        Provides the filesystem location where the keyword was found.
        """

    @property
    def lineno(self) -> int:
        pass

    @property
    def end_lineno(self) -> int:
        pass

    @property
    def col_offset(self) -> int:
        pass

    @property
    def end_col_offset(self) -> int:
        pass

    @property
    def library_name(self) -> Optional[str]:
        # If it's a library, this is the name of the library.
        pass

    @property
    def resource_name(self) -> Optional[str]:
        # If it's a resource, this is the basename of the resource without the extension.
        pass

    @property
    def library_alias(self) -> Optional[str]:
        pass

    # These are added if possible if there's some range to include the
    # full scope of the keyword. It should always include
    # the lineno/end_lineno range (so, it's a superset).

    @property
    def scope_lineno(self) -> Optional[int]:
        pass

    @property
    def scope_end_lineno(self) -> Optional[int]:
        pass

    @property
    def scope_col_offset(self) -> Optional[int]:
        pass

    @property
    def scope_end_col_offset(self) -> Optional[int]:
        pass


class IKeywordCollector(Protocol):
    def accepts(self, keyword_name: str) -> bool:
        """
        :param keyword_name:
            The name of the keyword to be accepted or not.
        :return bool:
            If the return is True, on_keyword(...) is called (otherwise it's not
            called).
        """

    def on_keyword(self, keyword_found: IKeywordFound):
        """
        :param IKeywordFound keyword_found:
        """

    def on_resolved_library(
        self,
        completion_context: "ICompletionContext",
        library_node: Optional[INode],
        library_doc: "ILibraryDoc",
    ):
        pass

    def on_unresolved_library(
        self,
        completion_context: "ICompletionContext",
        library_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        pass

    def on_unresolved_resource(
        self,
        completion_context: "ICompletionContext",
        resource_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        pass


class AbstractKeywordCollector:
    def on_resolved_library(
        self,
        completion_context: "ICompletionContext",
        library_node,
        library_doc: ILibraryDoc,
    ):
        pass

    def on_unresolved_library(
        self,
        completion_context: "ICompletionContext",
        library_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        pass

    def on_unresolved_resource(
        self,
        completion_context: "ICompletionContext",
        resource_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        pass


class IDefinition(Protocol):

    keyword_name: str = ""  # Can be empty if it's not found as a keyword.

    # Note: Could be None (i.e.: we found it in a library spec file which doesn't have the source).
    source: str = ""

    # Note: if we found it in a library spec file which doesn't have the lineno, it should be 0
    lineno: int = 0

    # Note: if we found it in a library spec file which doesn't have the lineno, it should be 0
    end_lineno: int = 0

    col_offset: int = 0

    end_col_offset: int = 0

    # These are added if possible if there's some range to include the
    # full scope (of a keyword, test, etc). It should always include
    # the lineno/end_lineno range (so, it's a superset).
    scope_lineno: Optional[int] = None
    scope_end_lineno: Optional[int] = None
    scope_col_offset: Optional[int] = None
    scope_end_col_offset: Optional[int] = None

    def hover_docs(self) -> MarkupContentTypedDict:
        pass


class IKeywordDefinition(IDefinition, Protocol):

    keyword_found: IKeywordFound


def cast_to_keyword_definition(definition: IDefinition) -> Optional[IKeywordDefinition]:
    if hasattr(definition, "keyword_found"):
        return typing.cast(IKeywordDefinition, definition)
    return None


class IBaseCompletionContext(Protocol):
    @property
    def monitor(self) -> Optional[IMonitor]:
        pass

    @property
    def workspace(self) -> IRobotWorkspace:
        pass

    @property
    def config(self) -> Optional[IConfig]:
        pass

    def check_cancelled(self) -> None:
        pass


class CompletionType(enum.Enum):
    regular = 1
    shell = 2


class LibraryDependencyInfo:
    def __init__(
        self,
        name: str,
        alias: Optional[str],
        builtin: bool,
        args: Optional[str],
        node: Optional[ILibraryImportNode],
    ):
        """
        :param builtin:
            Note that builtin should only be set == True if it's actually known
            that it's a builtin, otherwise it should be set to False (in which
            case it's computed internally if it's builtin or not).
        """
        self.name = name
        self.alias = alias
        self.builtin = builtin
        self.args = args
        self.node = node

    def to_dict(self):
        ret = {
            "name": self.name,
        }
        if self.alias:
            ret["alias"] = self.alias
        if self.builtin:
            ret["builtin"] = self.builtin
        if self.args:
            ret["args"] = self.args
        return ret


class ICompletionContextDependencyGraph(Protocol):
    def add_library_infos(
        self,
        doc_uri: str,
        library_infos: OrderedSet[LibraryDependencyInfo],
    ):
        pass

    def add_resource_infos(
        self,
        doc_uri: str,
        resource_imports_as_docs: Sequence[
            Tuple[IResourceImportNode, Optional[IRobotDocument]]
        ],
    ):
        pass

    def add_variable_infos(
        self,
        doc_uri: str,
        new_variable_imports: Sequence[
            Tuple[IVariableImportNode, Optional[IRobotDocument]]
        ],
    ):
        pass

    def get_root_doc(self) -> IRobotDocument:
        pass

    def iter_libraries(self, doc_uri: str) -> Iterator[LibraryDependencyInfo]:
        """
        Provides an iterator(doc_uri, library_dependency_infos)
        """

    def iter_all_libraries(self) -> Iterator[LibraryDependencyInfo]:
        pass

    def iter_resource_imports_with_docs(
        self, doc_uri: str
    ) -> Iterator[Tuple[IResourceImportNode, Optional[IRobotDocument]]]:
        pass

    def iter_all_resource_imports_with_docs(
        self,
    ) -> Iterator[Tuple[IResourceImportNode, Optional[IRobotDocument]]]:
        pass

    def iter_variable_imports_as_docs(
        self, doc_uri: str
    ) -> Iterator[Tuple[IVariableImportNode, Optional[IRobotDocument]]]:
        pass

    def iter_all_variable_imports_as_docs(
        self,
    ) -> Iterator[Tuple[IVariableImportNode, Optional[IRobotDocument]]]:
        pass

    def to_dict(self) -> dict:
        pass

    def do_invalidate_on_uri_change(self, uri: str) -> bool:
        pass


class IVariablesFromArgumentsFileLoader(Protocol):
    def get_variables(self) -> Tuple["IVariableFound", ...]:
        pass


class ICompletionContext(Protocol):
    def __init__(
        self,
        doc,
        line=Sentinel.SENTINEL,
        col=Sentinel.SENTINEL,
        workspace=None,
        config=None,
        memo=None,
        monitor: IMonitor = NULL,
        variables_from_arguments_files_loader: Sequence[
            IVariablesFromArgumentsFileLoader
        ] = (),
    ) -> None:
        pass

    def resolve_completion_item(
        self, data, completion_item: CompletionItemTypedDict, monaco: bool = False
    ) -> None:
        pass

    @property
    def variables_from_arguments_files_loader(
        self,
    ) -> Sequence[IVariablesFromArgumentsFileLoader]:
        pass

    @property
    def type(self) -> CompletionType:
        pass

    @property
    def monitor(self) -> IMonitor:
        pass

    def check_cancelled(self):
        pass

    def create_copy_with_selection(self, line: int, col: int) -> "ICompletionContext":
        pass

    def create_copy(self, doc: IRobotDocument) -> "ICompletionContext":
        pass

    @property
    def original_doc(self) -> IRobotDocument:
        pass

    @property
    def original_sel(self) -> Any:
        pass

    @property
    def doc(self) -> IRobotDocument:
        pass

    @property
    def sel(self) -> IDocumentSelection:
        pass

    @property
    def memo(self) -> Any:
        pass

    @property
    def config(self) -> Optional[IConfig]:
        pass

    @property
    def workspace(self) -> IRobotWorkspace:
        pass

    def get_type(self) -> Any:
        pass

    def get_ast(self) -> Any:
        pass

    def get_ast_current_section(self) -> Any:
        """
        :rtype: robot.parsing.model.blocks.Section|NoneType
        """

    def get_section(self, section_name: str) -> Any:
        """
        :rtype: robot_constants.Section
        """

    def get_accepted_section_header_words(self) -> List[str]:
        pass

    def get_current_section_name(self) -> Optional[str]:
        pass

    def get_current_token(self) -> Optional[TokenInfo]:
        pass

    def get_all_variables(self) -> Tuple[NodeInfo, ...]:
        pass

    def get_doc_normalized_var_name_to_var_found(self) -> Dict[str, "IVariableFound"]:
        pass

    def get_settings_normalized_var_name_to_var_found(
        self,
    ) -> Dict[str, "IVariableFound"]:
        pass

    def get_builtins_normalized_var_name_to_var_found(
        self, resolved
    ) -> Dict[str, "IVariableFound"]:
        pass

    def get_arguments_files_normalized_var_name_to_var_found(
        self,
    ) -> Dict[str, "IVariableFound"]:
        pass

    def get_current_variable(self, section=None) -> Optional[VarTokenInfo]:
        """
        Provides the current variable token. Note that it won't include '{' nor '}'.
        """

    def get_resource_import_as_doc(self, resource_import) -> Optional[IRobotDocument]:
        pass

    def get_variable_imports(self) -> Tuple[INode, ...]:
        pass

    def get_variable_import_as_doc(self, variables_import) -> Optional[IRobotDocument]:
        pass

    def get_current_keyword_definition(self) -> Optional[IKeywordDefinition]:
        pass

    def get_resource_imports(
        self,
    ) -> Tuple[IResourceImportNode, ...]:
        pass

    def get_resource_imports_as_docs(
        self,
    ) -> Tuple[Tuple[IResourceImportNode, Optional[IRobotDocument]], ...]:
        pass

    def get_resource_inits_as_docs(self) -> Tuple[IRobotDocument, ...]:
        pass

    def get_variable_imports_as_docs(
        self,
    ) -> Tuple[Tuple[IVariableImportNode, Optional[IRobotDocument]], ...]:
        pass

    def get_imported_libraries(self) -> Tuple[ILibraryImportNode, ...]:
        pass

    def token_value_resolving_variables(self, token: IRobotToken) -> str:
        pass

    def token_value_and_unresolved_resolving_variables(
        self, token: IRobotToken
    ) -> Tuple[str, Tuple[IRobotToken, ...]]:
        pass

    def get_current_keyword_definition_and_usage_info(
        self,
    ) -> Optional[Tuple[IKeywordDefinition, KeywordUsageInfo]]:
        pass

    def get_current_keyword_usage_info(
        self,
    ) -> Optional[KeywordUsageInfo]:
        pass

    def assign_documentation_resolve(
        self,
        completion_item: CompletionItemTypedDict,
        compute_documentation: Callable[[], MarkupContentTypedDict],
    ) -> None:
        pass

    def collect_dependency_graph(self) -> ICompletionContextDependencyGraph:
        pass

    def iter_dependency_and_init_resource_docs(
        self, dependency_graph
    ) -> Iterator[IRobotDocument]:
        pass


class VariableKind:
    VARIABLE = "Variable"
    BUILTIN = "Builtin Variable"
    ARGUMENT = "Argument"
    ENV_VARIABLE = "Environment Variable"
    SETTINGS = "Variable (settings)"
    PYTHON = "Variable (python)"
    YAML = "Variable (yaml)"
    ARGUMENTS_FILE = "Arguments file"


class IVariableFound(Protocol):
    """
    :ivar variable_name:
        This is the value that we should use when completing.
        It's the name of the variable without `${}` chars.

    :ivar variable_value:
        The value of the variable -- in general used to show information
        regarding that variable to the user.

    :ivar completion_context:
        This may be a new completion context, created when a new document is
        being analyzed (the variable was created for that completion context).

    :ivar source:
        Source where the variable was found.

    :ivar lineno:
        Line where it was found (0-based).
    """

    variable_name: str = ""
    variable_value: str = ""
    variable_kind: str = VariableKind.VARIABLE
    completion_context: Optional[ICompletionContext] = None

    @property
    def source(self) -> str:
        pass

    # Note: line/offsets 0-based.
    @property
    def lineno(self) -> int:
        pass

    @property
    def end_lineno(self) -> int:
        pass

    @property
    def col_offset(self) -> int:
        pass

    @property
    def end_col_offset(self) -> int:
        pass


class IVariablesCollector(Protocol):
    def accepts(self, variable_name: str) -> bool:
        """
        :param variable_name:
            The name of the variable (i.e.: ${some_var}).
        """

    def on_variable(self, variable_found: IVariableFound):
        pass

    def on_unresolved_variable_import(
        self,
        completion_context: "ICompletionContext",
        variable_import_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        pass


class AbstractVariablesCollector:
    def on_unresolved_variable_import(
        self,
        completion_context: "ICompletionContext",
        variable_import_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
    ):
        pass


class EvaluatableExpressionTypedDict(TypedDict):
    range: RangeTypedDict
    expression: Optional[str]


class IOnDependencyChanged(Protocol):
    def __call__(self, uri: str):
        pass



# Original work Copyright Fabio Zadrozny (EPL 1.0)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from functools import partial
import itertools
import os.path
import threading
from typing import Optional, List, Union, Tuple, Set
import typing

from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import BaseSchema
from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
    LaunchRequest,
    LaunchResponse,
    DisconnectRequest,
)
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_debug_adapter.base_launch_process_target import (
    IProtocolMessageCallable,
)
from robotframework_debug_adapter.constants import DEBUG
from robotframework_ls.options import DEFAULT_TIMEOUT


if typing.TYPE_CHECKING:
    from robotframework_debug_adapter.debug_adapter_comm import DebugAdapterComm

log = get_logger(__name__)


def _noop(*args, **kwargs):
    pass


def _read_stream(stream, on_line, category):
    try:
        while True:
            output = stream.readline()
            if len(output) == 0:
                log.debug("Finished reading stream: %s.\n" % (category,))
                break
            output = output.decode("utf-8", errors="replace")
            on_line(output, category)
    except:
        log.exception("Error")


def _notify_on_exited_pid(on_exit, pid):
    try:
        from robocorp_ls_core.basic import is_process_alive
        import time

        log.debug("Waiting for pid to exit (_notify_on_exited_pid).")

        while True:
            if not is_process_alive(pid):
                break

            time.sleep(0.2)
        log.debug("pid exited (_notify_on_exited_pid).")
        on_exit()
    except:
        log.exception("Error")


def find_init_robot_root(initial_target: str, up_to_dir: str) -> Optional[str]:
    """
    Find the root __init__.robot file which should be considered to create a
    suite based on the initial_target given.

    Returns the folder to be considered the root (which contains the __init__.robot).
    If an __init__.robot file is not found returns None.
    """
    check_dir: str

    if os.path.isdir(initial_target):
        check_dir = initial_target
    else:
        check_dir = os.path.dirname(initial_target)

    found: Optional[str] = None
    while True:
        if os.path.exists(os.path.join(check_dir, "__init__.robot")):
            found = check_dir

        if up_to_dir:
            if os.path.samefile(check_dir, up_to_dir):
                return found

        parent = os.path.dirname(check_dir)
        if not parent or parent == check_dir:
            return found

        check_dir = parent

    # Never gets here...


def _compute_env_filtering(env: dict, target_args: List[str]):
    import json

    env_filtering = {}
    if "RFLS_PRERUN_FILTER_TESTS" not in env:
        # If the RFLS_PRERUN_FILTER_TESTS env var is specified, it means that
        # the UI chose to have fine grained control of what needs to be run.
        # If that's not the case, we'll consider that we need to run anything
        # under the given path.
        #
        # Note that this is redundant in the case where makeSuite == False as
        # we'll just run the file/directories with everything in it (the user
        # can still filter with '-t' as needed in this case, but the suite will
        # only include what he added anyways and not more items as is the case
        # when makeSuite == True).
        include = [[path, "*"] for path in target_args]

        # We need to compute the filtering based on the target(s)
        env_filtering["RFLS_PRERUN_FILTER_TESTS"] = json.dumps(
            {"include": include, "exclude": []}
        )

    return env_filtering


def compute_cmd_line_and_env(
    run_robot_py: str,
    target: Union[str, List[str]],
    make_suite: bool,
    port: int,
    args: List[str],
    run_in_debug_mode: bool,
    cwd: str,
    suite_target: Optional[Union[str, List[str]]],
    env: dict,
) -> Tuple[List[str], dict]:
    """
    Note that cwd and target MUST be absolute at this point.
    """

    # This will be used if make_suite is not specified.
    target_args: List[str] = target if isinstance(target, list) else [target]
    new_target_args: List[str] = target_args[:]

    import sys

    new_env = env.copy()
    need_env_filtering = False

    suite_filter_args = []
    if make_suite:
        if suite_target:
            new_target_args = (
                suite_target if isinstance(suite_target, list) else [suite_target]
            )
            need_env_filtering = True
        else:
            # Ok, suite target hasn't been specified, so, we'll check if we need
            # to build a suite based on the `__init__.robot` files found in the
            # parent structure.
            found_roots: Set[Union[str, None]] = set()

            new_target_args = []
            for target in target_args:
                found = find_init_robot_root(target, cwd)
                found_roots.add(found)
                if not found:
                    if target not in new_target_args:
                        new_target_args.append(target)
                else:
                    if found not in new_target_args:
                        new_target_args.append(found)

            assert len(found_roots) > 0  # Even if it wasn't resolved, None is there ;)

            if len(found_roots) == 1:
                base_root = next(iter(found_roots))
                if base_root is None:
                    # I.e.: no __init__.robot is found anywhere, so, we
                    # don't need to do anything.
                    need_env_filtering = False
                else:
                    for target in target_args:
                        target_no_ext = os.path.splitext(target)[0]
                        relative = os.path.relpath(target_no_ext, base_root)
                        suite_to_filter = relative.replace("\\", "/").replace("/", ".")

                        suite_filter_args.append("--suite")
                        if suite_to_filter == ".":
                            suite_name = os.path.basename(base_root)
                        else:
                            suite_name = (
                                os.path.basename(base_root) + "." + suite_to_filter
                            )

                        if "__" in suite_name:
                            lst = []
                            for name in suite_name.split("."):
                                if name != "__init__":
                                    name = name.split("__", 1)[-1]
                                lst.append(name)
                            suite_name = ".".join(lst)
                        suite_filter_args.append(suite_name)

                    new_target_args = [base_root]

                    # We shouldn't need env filtering as we should've filtered
                    # for the suites we care about with --suite.
                    need_env_filtering = False

            else:
                # We don't have a common ancestor, so, just run all individually
                # (the new_target_args is already correct as is as we filled it
                # while computing the robot roots) and let the filtering take
                # place as needed.
                #
                # Note that this means all files will be loaded prior to the
                # filtering (unlike the case with a single root where we use
                # --suite which applies filtering prior to loading).
                need_env_filtering = True

    if need_env_filtering:
        new_env.update(_compute_env_filtering(env, target_args))

    if "RFLS_PRERUN_FILTER_TESTS" in new_env:
        found_filter = (
            "--prerunmodifier=robotframework_debug_adapter.prerun_modifiers.FilteringTestsSuiteVisitor"
            in args
        )
        if not found_filter:
            args.append(
                "--prerunmodifier=robotframework_debug_adapter.prerun_modifiers.FilteringTestsSuiteVisitor"
            )

    # Note: target must be the last parameter.
    cmdline = (
        [
            sys.executable,
            "-u",
            run_robot_py,
            "--port",
            str(port),
            "--debug" if run_in_debug_mode else "--no-debug",
        ]
        + args
        + suite_filter_args
        + new_target_args
    )
    return cmdline, new_env


class LaunchProcess(object):

    __slots__ = [
        "_valid",
        "_cmdline",
        "_terminal",
        "_popen",
        "_weak_debug_adapter_comm",
        "__weakref__",
        "_cwd",
        "_run_in_debug_mode",
        "_debug_adapter_robot_target_comm",
        "_debug_adapter_pydevd_target_comm",
        "_launch_response",
        "_next_seq",
        "_track_process_pid",
        "_env",
    ]

    def __init__(
        self,
        request: LaunchRequest,
        launch_response: LaunchResponse,
        debug_adapter_comm: DebugAdapterComm,
    ) -> None:
        self._init(request, launch_response, debug_adapter_comm)

    def _init(
        self,
        request: LaunchRequest,
        launch_response: LaunchResponse,
        debug_adapter_comm: DebugAdapterComm,
    ) -> None:
        import weakref
        from robotframework_debug_adapter.constants import VALID_TERMINAL_OPTIONS
        from robotframework_debug_adapter.constants import TERMINAL_INTEGRATED
        from robocorp_ls_core.basic import as_str
        import robocorp_ls_core
        from robotframework_debug_adapter.launch_process_robot_target_comm import (
            LaunchProcessDebugAdapterRobotTargetComm,
        )
        from robotframework_debug_adapter.launch_process_pydevd_comm import (
            LaunchProcessDebugAdapterPydevdComm,
        )

        self._weak_debug_adapter_comm = weakref.ref(debug_adapter_comm)
        self._valid = True
        self._cmdline: List[str] = []
        self._popen = None
        self._launch_response = launch_response
        self._next_seq = partial(next, itertools.count(0))
        self._track_process_pid = None

        def mark_invalid(message):
            launch_response.success = False
            launch_response.message = message
            self._valid = False

        target = request.arguments.kwargs.get("target")
        self._cwd = request.arguments.kwargs.get("cwd")
        self._terminal = request.arguments.kwargs.get("terminal", TERMINAL_INTEGRATED)
        args = request.arguments.kwargs.get("args") or []
        make_suite = request.arguments.kwargs.get("makeSuite", True)
        suite_target = request.arguments.kwargs.get("suiteTarget", "")
        args = [str(arg) for arg in args]

        env = {}
        request_env = request.arguments.kwargs.get("env")
        if isinstance(request_env, dict) and request_env:
            env.update(request_env)

        pythonpath = env.get("PYTHONPATH", "")
        pythonpath += (
            os.pathsep
            + os.path.dirname(os.path.dirname(__file__))
            + os.pathsep
            + os.path.dirname(os.path.dirname(robocorp_ls_core.__file__))
        )
        env["PYTHONPATH"] = pythonpath

        for key, value in os.environ.items():
            if "ROBOTFRAMEWORK" in key:
                env[key] = value

        env = dict(((as_str(key), as_str(value)) for (key, value) in env.items()))

        self._env = env

        self._run_in_debug_mode = not request.arguments.noDebug

        if self._terminal not in VALID_TERMINAL_OPTIONS:
            return mark_invalid(
                f"Invalid terminal option: {self._terminal} (must be one of: {VALID_TERMINAL_OPTIONS})"
            )

        try:
            if not target:
                return mark_invalid("target not provided in launch.")

            new_target = []

            if not isinstance(target, list):
                target = [target]

            for t in target:
                if not os.path.isabs(t):
                    if not self._cwd:
                        return mark_invalid(
                            f"Target: {t} is relative and cwd was not given."
                        )

                    t = os.path.abspath(os.path.join(self._cwd, t))
                else:
                    # This will also normalize
                    t = os.path.abspath(t)

                if not os.path.exists(t):
                    return mark_invalid(f"File: {t} does not exist.")

                new_target.append(t)

            target = new_target
        except Exception as e:
            msg = f"Error checking if target ({target}) exists:\n{e}"
            log.exception(msg)
            return mark_invalid(msg)

        try:
            if not self._cwd:
                if isinstance(target, list):
                    t = target[0]
                else:
                    t = target

                if os.path.isdir(t):
                    dirname = t
                else:
                    dirname = os.path.dirname(t)

                base = find_init_robot_root(dirname, "")
                if base is not None:
                    dirname = base

                self._cwd = dirname

            if not self._cwd:
                if not os.path.exists(self._cwd):
                    return mark_invalid(f"cwd specified does not exist: {self._cwd}")

            # make sure cwd is absolute
            self._cwd = os.path.abspath(self._cwd)
        except Exception as e:
            msg = f"Error checking if cwd ({self._cwd}) exists:\n{e}"
            log.exception(msg)
            return mark_invalid(msg)

        if DEBUG:
            log.debug("Run in debug mode: %s\n" % (self._run_in_debug_mode,))

        self._debug_adapter_robot_target_comm = (
            LaunchProcessDebugAdapterRobotTargetComm(debug_adapter_comm)
        )

        port, server_socket = self._debug_adapter_robot_target_comm.start_listening(
            2 if self._run_in_debug_mode else 1
        )

        self._debug_adapter_pydevd_target_comm = LaunchProcessDebugAdapterPydevdComm(
            debug_adapter_comm, server_socket
        )

        try:
            run_robot_py = os.path.join(
                os.path.dirname(__file__), "run_robot__main__.py"
            )
            if not os.path.exists(run_robot_py):
                return mark_invalid("File: %s does not exist." % (run_robot_py,))
        except Exception as e:
            msg = f"Error checking if run_robot__main__.py exists:\n{e}"
            log.exception(msg)
            return mark_invalid(msg)

        else:
            try:
                cmdline_and_env = compute_cmd_line_and_env(
                    run_robot_py,
                    target,
                    make_suite,
                    port,
                    args,
                    self._run_in_debug_mode,
                    self._cwd,
                    suite_target,
                    self._env,
                )

                self._cmdline, self._env = cmdline_and_env
            except Exception as e:
                msg = f"Error computing command line and environment:\n{e}"
                log.exception(msg)
                return mark_invalid(msg)

    @property
    def valid(self) -> bool:
        return self._valid

    @property
    def run_in_debug_mode(self) -> bool:
        return self._run_in_debug_mode

    def send_and_wait_for_configuration_done_request(self) -> bool:
        """
        :return: Whether the configuration done response was received.
        """
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            ConfigurationDoneRequest,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            ConfigurationDoneArguments,
        )

        event_robot = threading.Event()
        track_events = [event_robot]

        self._debug_adapter_robot_target_comm.write_to_robot_message(
            ConfigurationDoneRequest(ConfigurationDoneArguments()),
            on_response=lambda *args, **kwargs: event_robot.set(),
        )

        if self._run_in_debug_mode:
            event_pydevd = threading.Event()
            track_events.append(event_pydevd)
            self._debug_adapter_pydevd_target_comm.write_to_pydevd_message(
                ConfigurationDoneRequest(ConfigurationDoneArguments()),
                on_response=lambda *args, **kwargs: event_pydevd.set(),
            )

        log.debug(
            "Wating for configuration_done response for %s seconds."
            % (DEFAULT_TIMEOUT,)
        )
        ret = True
        for event in track_events:
            ret = ret and event.wait(DEFAULT_TIMEOUT)
            if not ret:
                break
        log.debug("Received configuration_done response: %s" % (ret,))
        return ret

    def resend_request_to_stopped_target(self, request: BaseSchema, target) -> None:
        if target is None:
            log.debug(f"Detected no paused backend to resend request: {request}")
            return

        if target is self._debug_adapter_pydevd_target_comm:
            self.resend_request_to_pydevd(request)
        elif target is self._debug_adapter_robot_target_comm:
            self.resend_request_to_robot(request)
        else:
            log.debug(
                f"Stopped target unexpected: {target} for resending request: {request}"
            )

    def resend_request_to_robot(self, request: BaseSchema) -> None:
        request_seq = request.seq

        def on_response(response_msg):
            response_msg.request_seq = request_seq
            debug_adapter_comm = self._weak_debug_adapter_comm()
            if debug_adapter_comm is not None:
                debug_adapter_comm.write_to_client_message(response_msg)
            else:
                log.debug(
                    "Command processor collected in resend request: %s" % (request,)
                )

        self._debug_adapter_robot_target_comm.write_to_robot_message(
            request, on_response
        )

    def resend_request_to_pydevd(self, request: BaseSchema) -> None:
        request_seq = request.seq

        def on_response(response_msg):
            response_msg.request_seq = request_seq
            debug_adapter_comm = self._weak_debug_adapter_comm()
            if debug_adapter_comm is not None:
                debug_adapter_comm.write_to_client_message(response_msg)
            else:
                log.debug(
                    "Command processor collected in resend request: %s" % (request,)
                )

        self._debug_adapter_pydevd_target_comm.write_to_pydevd_message(
            request, on_response
        )

    def write_to_robot_and_pydevd(self, request: BaseSchema):
        self._debug_adapter_robot_target_comm.write_to_robot_message(request)
        if self._run_in_debug_mode:
            self._debug_adapter_pydevd_target_comm.write_to_pydevd_message(request)

    def write_to_pydevd(
        self,
        request: BaseSchema,
        on_response: Optional[IProtocolMessageCallable] = None,
    ):
        if self._run_in_debug_mode:
            self._debug_adapter_pydevd_target_comm.write_to_pydevd_message(
                request, on_response=on_response
            )

    def launch(self):
        from robotframework_debug_adapter.constants import TERMINAL_NONE
        from robotframework_debug_adapter.constants import TERMINAL_EXTERNAL
        from robotframework_debug_adapter.constants import TERMINAL_INTEGRATED
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            RunInTerminalRequest,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            RunInTerminalRequestArguments,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import OutputEvent
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import OutputEventBody
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import InitializeRequest
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            InitializeRequestArguments,
        )

        # Note: using a weak-reference so that callbacks don't keep it alive
        weak_debug_adapter_comm = self._weak_debug_adapter_comm

        terminal = self._terminal
        if not weak_debug_adapter_comm().supports_run_in_terminal:
            # If the client doesn't support running in the terminal we fallback to using the debug console.
            terminal = TERMINAL_NONE

        threads = []
        if terminal == TERMINAL_NONE:
            import subprocess

            if DEBUG:
                log.debug(
                    "Launching in debug console (not in terminal): %s"
                    % (self._cmdline,)
                )

            env = os.environ.copy()
            env.update(self._env)
            self._popen = subprocess.Popen(
                self._cmdline,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                cwd=self._cwd,
                env=env,
            )

            def on_output(output, category):
                debug_adapter_comm = weak_debug_adapter_comm()
                if debug_adapter_comm is not None:
                    output_event = OutputEvent(
                        OutputEventBody(output, category=category)
                    )
                    debug_adapter_comm.write_to_client_message(output_event)

            stdout_stream_thread = threading.Thread(
                target=_read_stream,
                args=(self._popen.stdout, on_output, "stdout"),
                name="Read stdout",
            )
            stderr_stream_thread = threading.Thread(
                target=_read_stream,
                args=(self._popen.stderr, on_output, "stderr"),
                name="Read stderr",
            )

            threads.append(stdout_stream_thread)
            threads.append(stderr_stream_thread)

        elif terminal in (TERMINAL_INTEGRATED, TERMINAL_EXTERNAL):
            kind = terminal

            if DEBUG:
                log.debug('Launching in "%s" terminal: %s' % (kind, self._cmdline))

            debug_adapter_comm = weak_debug_adapter_comm()
            cmdline = self._cmdline
            if debug_adapter_comm is not None:
                env = self._env
                if env:
                    from robocorp_ls_core import run_with_env

                    cmdline, env = run_with_env.update_cmdline_and_env(cmdline, env)

                debug_adapter_comm.write_to_client_message(
                    RunInTerminalRequest(
                        RunInTerminalRequestArguments(
                            cwd=self._cwd, args=cmdline, kind=kind, env=env
                        )
                    )
                )

        for t in threads:
            t.daemon = True

        if not self._debug_adapter_robot_target_comm.wait_for_connection():
            launch_response = self._launch_response
            launch_response.success = False
            launch_response.message = (
                "Debug adapter timed out waiting for Robot connection."
            )
            self._valid = False

        if self._run_in_debug_mode:
            # Ok, at this point we connected to the robot target debug adapter,
            # now, let's connect to pydevd too (if in debug mode).
            self._debug_adapter_pydevd_target_comm.start()
            if not self._debug_adapter_pydevd_target_comm.wait_for_connection():
                launch_response = self._launch_response
                launch_response.success = False
                launch_response.message = (
                    "Debug adapter timed out waiting for pydevd connection."
                )
                self._valid = False

        initialize_request = InitializeRequest(
            InitializeRequestArguments("robot-launch-process-adapter")
        )
        self.write_to_robot_and_pydevd(initialize_request)

        # Note: only start listening stdout/stderr when connected.
        for t in threads:
            t.start()

        if not self._debug_adapter_robot_target_comm.wait_for_process_event():
            launch_response = self._launch_response
            launch_response.success = False
            launch_response.message = (
                "Debug adapter timed out waiting for process event."
            )
            self._valid = False
        else:
            self._track_process_pid = self._debug_adapter_robot_target_comm.get_pid()

    def after_launch_response_sent(self) -> None:
        if self._track_process_pid is None:
            log.debug("Unable to track if pid is alive (pid unavailable).")
            return
        t = threading.Thread(
            target=_notify_on_exited_pid,
            args=(
                self._debug_adapter_robot_target_comm.notify_exit,
                self._track_process_pid,
            ),
            name="Track PID alive",
        )
        t.daemon = True
        t.start()

    def disconnect(self, disconnect_request: DisconnectRequest) -> None:
        from robocorp_ls_core.basic import kill_process_and_subprocesses

        if self._popen is not None:
            if self._popen.returncode is None:
                kill_process_and_subprocesses(self._popen.pid)
        else:
            kill_process_and_subprocesses(self._track_process_pid)

    def send_to_stdin(self, expression):
        popen = self._popen
        if popen is not None:
            try:
                log.debug("Sending: %s to stdin." % (expression,))

                def write_to_stdin(popen, expression):
                    popen.stdin.write(expression)
                    if not expression.endswith("\r") and not expression.endswith("\n"):
                        popen.stdin.write("\n")
                    popen.stdin.flush()

                # Do it in a thread (in theory the OS could have that filled up and we would never complete
                # trying to write -- although probably a bit far fetched, let's code as if that could actually happen).
                t = threading.Thread(
                    target=write_to_stdin,
                    args=(popen, expression),
                    name="Send to STDIN",
                )
                t.daemon = True
                t.start()
            except:
                log.exception("Error writing: >>%s<< to stdin." % (expression,))



from robocorp_ls_core.robotframework_log import get_logger
from typing import Any

log = get_logger(__name__)


class _RegisteredCommand(object):
    def __init__(self, command_name, expected_return_cls):
        self.command_name = command_name
        self.expected_return_type = expected_return_cls
        self.func = None
        self.attr_in_language_server = None


class _CommandDispatcher(object):
    def __init__(self):
        self._command_name_to_registered_command = {}

    def __call__(self, command_name, expected_return_cls=None):
        """
        :param expected_return_type:
            If None, the return type is not checked.
        """
        if isinstance(command_name, str):
            self._curr_registered_command = _RegisteredCommand(
                command_name, expected_return_cls
            )
            return self
        else:
            self._curr_registered_command.func = command_name
            self._command_name_to_registered_command[
                self._curr_registered_command.command_name
            ] = self._curr_registered_command
            return self._curr_registered_command.func

    def dispatch(self, language_server, command_name, arguments) -> Any:
        try:
            registered_command = self._command_name_to_registered_command[command_name]
            func = registered_command.func
            attr_in_language_server = registered_command.attr_in_language_server
            target = language_server
            if attr_in_language_server:
                target = getattr(language_server, attr_in_language_server)

            if registered_command.expected_return_type is None:
                return func(target, *arguments)
            else:
                ret = func(target, *arguments)
                assert isinstance(ret, registered_command.expected_return_type)
                return ret

        except Exception as e:
            error_msg = f"Error in command: {command_name} with args: {arguments}.\n{e}"
            log.exception(error_msg)
            return {"success": False, "message": error_msg, "result": None}

    def register_sub_command_dispatcher(self, sub_dispatcher):
        self._command_name_to_registered_command.update(
            sub_dispatcher._command_name_to_registered_command
        )


class _SubCommandDispatcher(object):
    def __init__(self, attr_in_language_server):
        self._command_name_to_registered_command = {}
        self.attr_in_language_server = attr_in_language_server

    def __call__(self, command_name, expected_return_cls=None):
        """
        :param expected_return_type:
            If None, the return type is not checked.
        """
        if isinstance(command_name, str):
            self._curr_registered_command = _RegisteredCommand(
                command_name, expected_return_cls
            )
            return self
        else:
            self._curr_registered_command.func = command_name
            self._curr_registered_command.attr_in_language_server = (
                self.attr_in_language_server
            )
            self._command_name_to_registered_command[
                self._curr_registered_command.command_name
            ] = self._curr_registered_command
            return self._curr_registered_command.func



from _pydevd_bundle.pydevd_constants import DebugInfoHolder, \
    get_global_debugger, GetGlobalDebugger, set_global_debugger  # Keep for backward compatibility @UnusedImport
from _pydevd_bundle.pydevd_utils import quote_smart as quote, to_string
from _pydevd_bundle.pydevd_comm_constants import ID_TO_MEANING, CMD_EXIT
from _pydevd_bundle.pydevd_constants import HTTP_PROTOCOL, HTTP_JSON_PROTOCOL, \
    get_protocol, IS_JYTHON, ForkSafeLock
import json
from _pydev_bundle import pydev_log


class _BaseNetCommand(object):

    # Command id. Should be set in instance.
    id = -1

    # Dict representation of the command to be set in instance. Only set for json commands.
    as_dict = None

    def send(self, *args, **kwargs):
        pass


class _NullNetCommand(_BaseNetCommand):
    pass


class _NullExitCommand(_NullNetCommand):

    id = CMD_EXIT


# Constant meant to be passed to the writer when the command is meant to be ignored.
NULL_NET_COMMAND = _NullNetCommand()

# Exit command -- only internal (we don't want/need to send this to the IDE).
NULL_EXIT_COMMAND = _NullExitCommand()


class NetCommand(_BaseNetCommand):
    """
    Commands received/sent over the network.

    Command can represent command received from the debugger,
    or one to be sent by daemon.
    """
    next_seq = 0  # sequence numbers

    _showing_debug_info = 0
    _show_debug_info_lock = ForkSafeLock(rlock=True)

    def __init__(self, cmd_id, seq, text, is_json=False):
        """
        If sequence is 0, new sequence will be generated (otherwise, this was the response
        to a command from the client).
        """
        protocol = get_protocol()
        self.id = cmd_id
        if seq == 0:
            NetCommand.next_seq += 2
            seq = NetCommand.next_seq

        self.seq = seq

        if is_json:
            if hasattr(text, 'to_dict'):
                as_dict = text.to_dict(update_ids_to_dap=True)
            else:
                assert isinstance(text, dict)
                as_dict = text
            as_dict['pydevd_cmd_id'] = cmd_id
            as_dict['seq'] = seq
            self.as_dict = as_dict
            text = json.dumps(as_dict)

        assert isinstance(text, str)

        if DebugInfoHolder.DEBUG_TRACE_LEVEL >= 1:
            self._show_debug_info(cmd_id, seq, text)

        if is_json:
            msg = text
        else:
            if protocol not in (HTTP_PROTOCOL, HTTP_JSON_PROTOCOL):
                encoded = quote(to_string(text), '/<>_=" \t')
                msg = '%s\t%s\t%s\n' % (cmd_id, seq, encoded)

            else:
                msg = '%s\t%s\t%s' % (cmd_id, seq, text)

        if isinstance(msg, str):
            msg = msg.encode('utf-8')

        assert isinstance(msg, bytes)
        as_bytes = msg
        self._as_bytes = as_bytes

    def send(self, sock):
        as_bytes = self._as_bytes
        try:
            if get_protocol() in (HTTP_PROTOCOL, HTTP_JSON_PROTOCOL):
                sock.sendall(('Content-Length: %s\r\n\r\n' % len(as_bytes)).encode('ascii'))
            sock.sendall(as_bytes)
        except:
            if IS_JYTHON:
                # Ignore errors in sock.sendall in Jython (seems to be common for Jython to
                # give spurious exceptions at interpreter shutdown here).
                pass
            else:
                raise

    @classmethod
    def _show_debug_info(cls, cmd_id, seq, text):
        with cls._show_debug_info_lock:
            # Only one thread each time (rlock).
            if cls._showing_debug_info:
                # avoid recursing in the same thread (just printing could create
                # a new command when redirecting output).
                return

            cls._showing_debug_info += 1
            try:
                out_message = 'sending cmd (%s) --> ' % (get_protocol(),)
                out_message += "%20s" % ID_TO_MEANING.get(str(cmd_id), 'UNKNOWN')
                out_message += ' '
                out_message += text.replace('\n', ' ')
                try:
                    pydev_log.critical('%s\n', out_message)
                except:
                    pass
            finally:
                cls._showing_debug_info -= 1




import time


def method2():
    i = 1


def method():

    for i in range(200000):
        method2()

        if False:
            # Unreachable breakpoint here
            pass


def caller():
    start_time = time.time()
    method()
    print('TotalTime>>%s<<' % (time.time() - start_time,))


if __name__ == '__main__':
    import sys
    if '--regular-trace' in sys.argv:

        def trace_dispatch(frame, event, arg):
            return trace_dispatch

        sys.settrace(trace_dispatch)

    caller()  # Initial breakpoint for a step-over here
    print('TEST SUCEEDED')
